/*
sor.cu (CUDA + MPI)

SOR method (red-black)
*/

#ifdef _MPI
#include <mpi.h>
#endif

#include "ost.h"
#include "ost_mpi.h"
#include "ost_cuda.h"
#include "ost_prototype.h"
#include "../cuda/reduction_sum.cu"

// GPU/CPU
__host__ __device__
static real_t update(
	int i, int j, int k, int64_t ni, int64_t nj, int64_t nk, int64_t n,
	real_t *rxp, real_t *ryp, real_t *rzp, real_t *rxm, real_t *rym, real_t *rzm,
	real_t *v, id_t *idvolt, id_t *idepsr, real_t *epsr, real_t omega)
{
	const real_t e = epsr[idepsr[n]];

	const real_t axp = (e + epsr[idepsr[n + ni]]) * rxp[i];
	const real_t axm = (e + epsr[idepsr[n - ni]]) * rxm[i];
	const real_t ayp = (e + epsr[idepsr[n + nj]]) * ryp[j];
	const real_t aym = (e + epsr[idepsr[n - nj]]) * rym[j];
	const real_t azp = (e + epsr[idepsr[n + nk]]) * rzp[k];
	const real_t azm = (e + epsr[idepsr[n - nk]]) * rzm[k];

	const real_t res = (idvolt[n] ? 0 : 1) * omega * ((
		axp * v[n + ni] +
		axm * v[n - ni] +
		ayp * v[n + nj] +
		aym * v[n - nj] +
		azp * v[n + nk] +
		azm * v[n - nk]) / (axp + axm + ayp + aym + azp + azm) - v[n]);

	v[n] += res;

	return (res * res);
}

// GPU
__global__
static void update_gpu(int oe,
	int imin, int imax, int jmin, int jmax, int kmin, int kmax, int imin2, int jmin2, int kmin2,
	int64_t ni, int64_t nj, int64_t nk,
	real_t *rxp, real_t *ryp, real_t *rzp, real_t *rxm, real_t *rym, real_t *rzm,
	real_t *v, id_t *idvolt, id_t *idepsr, real_t *epsr, real_t omega, real_t *d_res2)
{
	extern __shared__ real_t sm[];

	const int i = imin + blockIdx.z;
	const int j = jmin + threadIdx.y + (blockIdx.y * blockDim.y);
	const int k = kmin + threadIdx.x + (blockIdx.x * blockDim.x);

	const int nthread = blockDim.x * blockDim.y;
	const int tid = threadIdx.x + (threadIdx.y * blockDim.x);
	const int bid = blockIdx.x + (blockIdx.y * gridDim.x) + (blockIdx.z * gridDim.x * gridDim.y);

	real_t res2 = 0;

	if (i <= imax) {
	if (j <= jmax) {
	if (k <= kmax) {
		if ((i + j + k) % 2 == oe) {
			const int64_t n = ((i - imin2 + 1) * ni) + ((j - jmin2 + 1) * nj) + ((k - kmin2 + 1) * nk);
			res2 = update(
				i, j, k, ni, nj, nk, n,
				rxp, ryp, rzp, rxm, rym, rzm,
				v, idvolt, idepsr, epsr, omega);
		}
	}
	}
	}
	sm[tid] = res2;

	reduction_sum(tid, nthread, sm, &d_res2[bid]);
}

// CPU
static real_t update_cpu(int oe,
	int imin, int imax, int jmin, int jmax, int kmin, int kmax, int imin2, int jmin2, int kmin2,
	int64_t ni, int64_t nj, int64_t nk,
	real_t *rxp, real_t *ryp, real_t *rzp, real_t *rxm, real_t *rym, real_t *rzm,
	real_t *v, id_t *idvolt, id_t *idepsr, real_t *epsr, real_t omega)
{
	real_t res2 = 0;

	for (int i = imin; i <= imax; i++) {
	for (int j = jmin; j <= jmax; j++) {
	for (int k = kmin; k <= kmax; k++) {
		if ((i + j + k) % 2 == oe) {
			const int64_t n = ((i - imin2 + 1) * ni) + ((j - jmin2 + 1) * nj) + ((k - kmin2 + 1) * nk);
			res2 += update(
				i, j, k, ni, nj, nk, n,
				rxp, ryp, rzp, rxm, rym, rzm,
				v, idvolt, idepsr, epsr, omega);
		}
	}
	}
	}

	return res2;
}

static void memory_alloc();
static void memory_copy();
static void memory_free();
static void copy_comm_boundary(real_t *, real_t *);

void sor(int vector, FILE *fp)
{
	const char msg1[] = "  --- converged ---";
	const char msg2[] = "  --- diverged ---";
	const char msg3[] = "  --- not converged ---";
	const char fmt[] = "   %7d    %.7f";
	char str[BUFSIZ];

	vector = vector; // dummy
	const int64_t num = (iMax2 - iMin2 + 3) * (jMax2 - jMin2 + 3) * (kMax2 - kMin2 + 3);

	// V
	V = (real_t *)malloc(num * sizeof(real_t));
	memset(V, 0, num * sizeof(real_t));

	// residual
	const size_t r_size = (Solver.maxiter + 1) * sizeof(double);
	Residual = (double *)malloc(r_size);
	memset(Residual, 0, r_size);

	// index
	const int64_t ni = D(1, 0, 0) - D(0, 0, 0);
	const int64_t nj = D(0, 1, 0) - D(0, 0, 0);
	const int64_t nk = D(0, 0, 1) - D(0, 0, 0);

	// MPI buffer
	const size_t b_size = (Ny + 3) * (Nz + 3) * sizeof(real_t);
	real_t *sendbuf = (real_t *)malloc(b_size);
	real_t *recvbuf = (real_t *)malloc(b_size);

	// electrode
	electrode(iMin2, iMax2, jMin2, jMax2, kMin2, kMax2);

	// initial field
	real_t vmin, vmax;
	initial(&vmin, &vmax, iMin2, iMax2, jMin2, jMax2, kMin2, kMax2);

	// alloc device memory
	if (GPU) memory_alloc();

	// execution configuration
	dim3 block = dim3(16, 16, 1);
	dim3 grid = dim3(CEIL(kMax - kMin + 1, block.x), CEIL(jMax - jMin + 1, block.y), iMax - iMin + 1);

	// residual array
	size_t r2_size = 0;
	real_t *d_res2 = NULL, *h_res2 = NULL;
	if (GPU) {
		r2_size = grid.x * grid.y * grid.z * sizeof(real_t);
		cuda_malloc(GPU, UM, (void **)&d_res2, r2_size);
		h_res2 = (real_t *)malloc(r2_size);
		memset(h_res2, 0, r2_size);
	}

	// iteration
	int iter;
	for (iter = 0; iter <= Solver.maxiter; iter++) {
		// update
		real_t res2 = 0;
		for (int oe = 0; oe < 2; oe++) {
			if (GPU) {
				// GPU
				int sm_size = block.x * block.y * sizeof(real_t);
				update_gpu<<<grid, block, sm_size>>>(oe,
					iMin, iMax, jMin, jMax, kMin, kMax, iMin2, jMin2, kMin2,
					ni, nj, nk,
					d_RXp, d_RYp, d_RZp, d_RXm, d_RYm, d_RZm,
					d_V, d_idVolt, d_idEpsr, d_Epsr, (real_t)Solver.omega, d_res2);
				cudaMemcpy(h_res2, d_res2, r2_size, cudaMemcpyDeviceToHost);
				for (int n = 0; n < r2_size / sizeof(real_t); n++) {
					res2 += h_res2[n];
				}

			}
			else {
				// CPU
				res2 += update_cpu(oe,
					iMin, iMax, jMin, jMax, kMin, kMax, iMin2, jMin2, kMin2,
					ni, nj, nk,
					RXp, RYp, RZp, RXm, RYm, RZm,
					V, idVolt, idEpsr, Epsr, (real_t)Solver.omega);
			}

			// share boundary
			if (commSize > 1) {
				copy_comm_boundary(sendbuf, recvbuf);
			}
		}

		// residual
		res2 = comm_sum(res2);
		Residual[iter] = sqrt(res2 / ((double)(Nx + 1) * (Ny + 1) * (Nz + 1)));

		// monitor
		if ((iter % Solver.nout) == 0 || (Residual[iter] < Solver.converg)) {
			sprintf(str, fmt, iter, Residual[iter]);
			if (commRank == 0) monitor1(fp, str);
		}

		// converged / diverged / not converged
		if (Residual[iter] < Solver.converg) {
			if (commRank == 0) monitor1(fp, msg1);
			break;
		}
		if (Residual[iter] > 1e10) {
			if (commRank == 0) monitor1(fp, msg2);
			break;
		}
		if (iter == Solver.maxiter) {
			if (commRank == 0) monitor1(fp, msg3);
			break;
		}
	}

	// iterations
	NumIter = iter + 1;

	// copy device memory
	if (GPU) {
		memory_copy();
	}

	// free device memory
	if (GPU) {
		memory_free();
	}

	// free
	if (GPU) {
		cuda_free(GPU, d_res2);
		free(h_res2);
	}

	// final
	final(vmin, vmax, iMin2, iMax2, jMin2, jMax2, kMin2, kMax2);

	// edge
	edge(iMin2, iMax2, jMin2, jMax2, kMin2, kMax2);

	// free
	free(sendbuf);
	free(recvbuf);
}

static void memory_alloc()
{
	int64_t num = (int64_t)(iMax2 - iMin2 + 3) * (jMax2 - jMin2 + 3) * (kMax2 - kMin2 + 3);

	cuda_malloc(GPU, UM, (void **)&d_V,           num * sizeof(real_t));
	cuda_malloc(GPU, UM, (void **)&d_idVolt,      num * sizeof(id_t));
	cuda_malloc(GPU, UM, (void **)&d_idEpsr,      num * sizeof(id_t));
	cuda_malloc(GPU, UM, (void **)&d_RXp,    (Nx + 1) * sizeof(real_t));
	cuda_malloc(GPU, UM, (void **)&d_RYp,    (Ny + 1) * sizeof(real_t));
	cuda_malloc(GPU, UM, (void **)&d_RZp,    (Nz + 1) * sizeof(real_t));
	cuda_malloc(GPU, UM, (void **)&d_RXm,    (Nx + 1) * sizeof(real_t));
	cuda_malloc(GPU, UM, (void **)&d_RYm,    (Ny + 1) * sizeof(real_t));
	cuda_malloc(GPU, UM, (void **)&d_RZm,    (Nz + 1) * sizeof(real_t));
	cuda_malloc(GPU, UM, (void **)&d_Epsr,        256 * sizeof(real_t));

	cuda_memcpy(GPU, d_V,      V,           num * sizeof(real_t), cudaMemcpyHostToDevice);
	cuda_memcpy(GPU, d_idVolt, idVolt,      num * sizeof(id_t),   cudaMemcpyHostToDevice);
	cuda_memcpy(GPU, d_idEpsr, idEpsr,      num * sizeof(id_t),   cudaMemcpyHostToDevice);
	cuda_memcpy(GPU, d_RXp,    RXp,    (Nx + 1) * sizeof(real_t), cudaMemcpyHostToDevice);
	cuda_memcpy(GPU, d_RYp,    RYp,    (Ny + 1) * sizeof(real_t), cudaMemcpyHostToDevice);
	cuda_memcpy(GPU, d_RZp,    RZp,    (Nz + 1) * sizeof(real_t), cudaMemcpyHostToDevice);
	cuda_memcpy(GPU, d_RXm,    RXm,    (Nx + 1) * sizeof(real_t), cudaMemcpyHostToDevice);
	cuda_memcpy(GPU, d_RYm,    RYm,    (Ny + 1) * sizeof(real_t), cudaMemcpyHostToDevice);
	cuda_memcpy(GPU, d_RZm,    RZm,    (Nz + 1) * sizeof(real_t), cudaMemcpyHostToDevice);
	cuda_memcpy(GPU, d_Epsr,   Epsr,        256 * sizeof(real_t), cudaMemcpyHostToDevice);
}

static void memory_copy()
{
	int64_t num = (int64_t)(iMax2 - iMin2 + 3) * (jMax2 - jMin2 + 3) * (kMax2 - kMin2 + 3);

	cuda_memcpy(GPU, V, d_V, num * sizeof(real_t), cudaMemcpyDeviceToHost);
}

static void memory_free()
{
	cuda_free(GPU, d_V);
	cuda_free(GPU, d_idVolt);
	cuda_free(GPU, d_idEpsr);
	cuda_free(GPU, d_RXp);
	cuda_free(GPU, d_RYp);
	cuda_free(GPU, d_RZp);
	cuda_free(GPU, d_RXm);
	cuda_free(GPU, d_RYm);
	cuda_free(GPU, d_RZm);
	cuda_free(GPU, d_Epsr);
}

// share boundary : copy D2H + comm + copy H2D
void copy_comm_boundary(real_t *sendbuf, real_t *recvbuf)
{
#ifdef _MPI
	MPI_Status status;
	const int tag = 0;
	const int count = (Ny + 3) * (Nz + 3);
	const size_t size = count * sizeof(real_t);
	real_t *ptr = GPU ? d_V : V;

	// -X boundary
	if (commRank > 0) {
		// copy to buffer
		cuda_memcpy(GPU, sendbuf, &ptr[D(1, -1, -1)], size, cudaMemcpyDeviceToHost);

		// MPI
		int dst = commRank - 1;
		MPI_Sendrecv(sendbuf, count, MPI_REAL_T, dst, tag,
		             recvbuf, count, MPI_REAL_T, dst, tag, MPI_COMM_WORLD, &status);

		// copy from buffer
		cuda_memcpy(GPU, &ptr[D(0, -1, -1)], recvbuf, size, cudaMemcpyHostToDevice);
	}

	// +X boundary
	if (commRank < commSize - 1) {
		// copy to buffer
		cuda_memcpy(GPU, sendbuf, &ptr[D(iMax2 - iMin2 - 1, -1, -1)], size, cudaMemcpyDeviceToHost);

		// MPI
		int dst = commRank + 1;
		MPI_Sendrecv(sendbuf, count, MPI_REAL_T, dst, tag,
		             recvbuf, count, MPI_REAL_T, dst, tag, MPI_COMM_WORLD, &status);

		// copy from buffer
		cuda_memcpy(GPU, &ptr[D(iMax2 - iMin2, -1, -1)], recvbuf, size, cudaMemcpyHostToDevice);
	}
#endif
}
