% input data
function [Parm, Mesh, Geom, err] = input_data(fid)
	err = 0;

    % initialize
	Geom(1) = struct( ...
		'type', 0, ...        % = 1/2 : volt/epsr
		'pid', 0, ...         % = 1... : property index
		'shape', 0, ...       % = 1,2,... : shape
		'pos', zeros(1, 8));  % position (6 or 8)

	Parm.title = "";
	Parm.volt = [];
	Parm.epsr = [];
	Parm.solver = [];
	Parm.plot3dgeom = 0;

	xn = [];
	yn = [];
	zn = [];

	nvolt = 0;
	nepsr = 0;
	ngeom = 0;

	% header
    tline = fgetl(fid);
    d = strsplit(strtrim(tline));
    if (d{1} ~= "OpenSTF") || (length(d) < 3) 
        disp('Not OpenSTF data.');
        return
    end
    %version = (100 * str2double(d{2})) + str2double(d{3});

    % body
	while ~feof(fid)
        tline = fgetl(fid); %disp(tline);
        d = strsplit(strtrim(tline));  % cell array
        
        if (startsWith(d{1}, "#"))
            continue;
        elseif (strcmp(d{1}, "end"))
            break;
        elseif (length(d) < 3)
            continue;
        elseif (d{2} ~= "=")
            continue;
        end

		% parameter array
        p = zeros(1, length(d) - 2);
		for i = 3 : length(d)
            p(i - 2) = str2double(d{i});
		end

        switch d{1}
            case "title"
                Parm.title = tline(length('title = ') + 1 : end);
			case "xmesh"
				xn = makemesh(p);
			case "ymesh"
				yn = makemesh(p);
			case "zmesh"
				zn = makemesh(p);
			case "volt"
				nvolt = nvolt + 1;
				Parm.volt(nvolt) = p(1);
			case "epsr"
				nepsr = nepsr + 1;
				Parm.epsr(nepsr) = p(1);
            case "geometry"
                if (length(p) > 8)
					ngeom = ngeom + 1;
					Geom(ngeom).type  = p(1);
					Geom(ngeom).pid   = p(2);
					Geom(ngeom).shape = p(3);
					Geom(ngeom).pos   = p(4:end);
                end
			case "solver"
				if (length(p) > 3)
					Parm.solver = p(1:4);
				end
			case "plot3dgeom"
				Parm.plot3dgeom = p(1);  % new
			otherwise
        end
	end

	fclose(fid);

	% error check
	if isempty(xn)
		err = 1;
		disp("*** No xmesh");
	end
	if isempty(yn)
		err = 1;
		disp("*** No ymesh");
	end
	if isempty(zn)
		err = 1;
		disp("*** No zmesh");
	end
	if (min(diff(xn)) < eps)
		err = 1;
		disp("*** zero interval in xmesh");
	end
	if (min(diff(yn)) < eps)
		err = 1;
		disp("*** zero interval in ymesh");
	end
	if (min(diff(zn)) < eps)
		err = 1;
		disp("*** zero interval in zmesh");
	end
	if (nvolt < 2)
		err = 1;
		disp("*** No. of volt < 2");
	end
	if (ngeom < 2)
		err = 1;
		disp("*** No. of geometry < 2")
	end
	if any(isnan(Parm.solver))
		err = 1;
		disp("*** Not a number in solver")
	end

	% mesh factors
	Mesh = meshfactor(xn, yn, zn);

	% misc
	C = 2.99792458e8;
	Parm.eps0 = 1 / (pi * 4e-7 * C^2);
end

% make mesh
function mesh = makemesh(region)
	mesh = [];
	if (mod(length(region), 2) == 0)
		region(end) = [];  % odd
	end
	nregion = (length(region) - 1) / 2;
	if (nregion <= 0)
		return  % no data
	end
	if (region(1) > region(end))
		region = flip(region);  % ascending order
	end

	% make mesh
	nmesh = 1;
	for ir = 1 : nregion
		if (region(2 * ir) < 1)
			region(2 * ir) = 1;
		end
		nmesh = nmesh + region(2 * ir);
	end
	mesh = zeros(1, nmesh);
	mesh(1) = region(1);
	id = 1;
	for ir = 1 : nregion
		r0 = region(2 * ir - 1);
		r1 = region(2 * ir + 1);
		div = region(2 * ir);
		dr = (r1 - r0) / div;
		for m = 1 : div
			id = id + 1;
			mesh(id) = r0 + (m * dr);
		end
	end
end

% make mesh and mesh factors
function Mesh = meshfactor(xn, yn, zn)
	nx = length(xn);
	ny = length(yn);
	nz = length(zn);
	Mesh.xn = xn;
	Mesh.yn = yn;
	Mesh.zn = zn;
	Mesh.nx = nx;
	Mesh.ny = ny;
	Mesh.nz = nz;

	% center
	Mesh.dxc = zeros(1, nx);
	Mesh.dyc = zeros(1, ny);
	Mesh.dzc = zeros(1, nz);
	for i = 2 : nx - 1
		Mesh.dxc(i) = (xn(i + 1) - xn(i - 1)) / 2;
	end
	for j = 2 : ny - 1
		Mesh.dyc(j) = (yn(j + 1) - yn(j - 1)) / 2;
	end
	for k = 2 : nz - 1
		Mesh.dzc(k) = (zn(k + 1) - zn(k - 1)) / 2;
	end
	%{
	Mesh.dxc(1) = (xn(2) - xn(1)) / 2;
	Mesh.dyc(1) = (yn(2) - yn(1)) / 2;
	Mesh.dzc(1) = (zn(2) - zn(1)) / 2;
	Mesh.dxc(nx) = (xn(nx) - xn(nx - 1)) / 2;
	Mesh.dyc(ny) = (yn(ny) - yn(ny - 1)) / 2;
	Mesh.dzc(nz) = (zn(nz) - zn(nz - 1)) / 2;
	%}
	Mesh.dxc(1) = xn(2) - xn(1);
	Mesh.dyc(1) = yn(2) - yn(1);
	Mesh.dzc(1) = zn(2) - zn(1);
	Mesh.dxc(nx) = xn(nx) - xn(nx - 1);
	Mesh.dyc(ny) = yn(ny) - yn(ny - 1);
	Mesh.dzc(nz) = zn(nz) - zn(nz - 1);

	% half plus
	Mesh.dxp = zeros(1, nx);
	Mesh.dyp = zeros(1, ny);
	Mesh.dzp = zeros(1, nz);
	for i = 1 : nx - 1
		Mesh.dxp(i) = xn(i + 1) - xn(i);
	end
	for j = 1 : ny - 1
		Mesh.dyp(j) = yn(j + 1) - yn(j);
	end
	for k = 1 : nz - 1
		Mesh.dzp(k) = zn(k + 1) - zn(k);
	end
	Mesh.dxp(nx) = Mesh.dxp(nx - 1);
	Mesh.dyp(ny) = Mesh.dyp(ny - 1);
	Mesh.dzp(nz) = Mesh.dzp(nz - 1);

	% half minus
	Mesh.dxm = zeros(1, nx);
	Mesh.dym = zeros(1, ny);
	Mesh.dzm = zeros(1, nz);
	for i = 2 : nx
		Mesh.dxm(i) = xn(i) - xn(i - 1);
	end
	for j = 2 : ny
		Mesh.dym(j) = yn(j) - yn(j - 1);
	end
	for k = 2 : nz
		Mesh.dzm(k) = zn(k) - zn(k - 1);
	end
	Mesh.dxm(1) = Mesh.dxm(2);
	Mesh.dym(1) = Mesh.dym(2);
	Mesh.dzm(1) = Mesh.dzm(2);

	% finite difference factor
	Mesh.rxp = 1 ./ (Mesh.dxc .* Mesh.dxp);
	Mesh.rxm = 1 ./ (Mesh.dxc .* Mesh.dxm);
	Mesh.ryp = 1 ./ (Mesh.dyc .* Mesh.dyp);
	Mesh.rym = 1 ./ (Mesh.dyc .* Mesh.dym);
	Mesh.rzp = 1 ./ (Mesh.dzc .* Mesh.dzp);
	Mesh.rzm = 1 ./ (Mesh.dzc .* Mesh.dzm);
end
