% solve V
function [V, Residual, Niter] = solve(Parm, Mesh, Geom, idVolt, Epsr, fid)
	% red-black flag
	rb = true;

	% mesh
	nx = Mesh.nx;
	ny = Mesh.ny;
	nz = Mesh.nz;

	% iteration parameters
	omega   = Parm.solver(1);
	maxiter = Parm.solver(2);
	nout    = Parm.solver(3);
	tol     = Parm.solver(4);

	% initial value and scaling
	[V, vscale] = initial(Parm, Geom, idVolt, nx, ny, nz);

	% iteration
	Residual = zeros(1, maxiter);
	for iter = 1 : maxiter
		% update V
		if (rb == true)
			[V, res20] = update_rb(0, V, Mesh, idVolt, Epsr, omega);
			[V, res21] = update_rb(1, V, Mesh, idVolt, Epsr, omega);
			res2 = res20 + res21;
		else
			[V, res2] = update(V, Mesh, idVolt, Epsr, omega);
		end

		% boundary condition
		V = bc(V, idVolt, nx, ny, nz);

		% residual
		Residual(iter) = sqrt(res2 / ((nx - 2) * (ny - 2) * (nz - 2)));

		% monitor
		if (mod(iter, nout) == 0) || (iter == 1) || (Residual(iter) < tol)
			str = sprintf('    %6d    %.7f', iter, Residual(iter));
			monitor(1, fid, str);
		end

		% converged
		if (Residual(iter) < tol)
			monitor(1, fid, "  --- converged.");
			break;
		end
		% diverged
		if (Residual(iter) > 1e10)
			monitor(1, fid, "  --- diverged.");
			break;
		end
		% not converged
		if (iter == maxiter)
			monitor(1, fid, "  --- not converged.");
			break;
		end
	end

	% rescaling
	V = final(V, vscale, ny, nz);
	
	% iteration times
	Niter = iter;
end

% SOR
function [V, res2] = update(V, Mesh, idVolt, Epsr, omega)
	res2 = 0;

	% mesh
	nx = Mesh.nx;
	ny = Mesh.ny;
	nz = Mesh.nz;
	rxp = Mesh.rxp;
	rxm = Mesh.rxm;
	ryp = Mesh.ryp;
	rym = Mesh.rym;
	rzp = Mesh.rzp;
	rzm = Mesh.rzm;

	for k = 2 : nz - 1
	for j = 2 : ny - 1
	for i = 2 : nx - 1
		%if (idVolt(i, j, k) == 0)
			e0 = Epsr(i, j, k);
	
			axp = (e0 + Epsr(i + 1, j,     k    )) * rxp(i);
			axm = (e0 + Epsr(i - 1, j,     k    )) * rxm(i);
			ayp = (e0 + Epsr(i,     j + 1, k    )) * ryp(j);
			aym = (e0 + Epsr(i,     j - 1, k    )) * rym(j);
			azp = (e0 + Epsr(i,     j,     k + 1)) * rzp(k);
			azm = (e0 + Epsr(i,     j,     k - 1)) * rzm(k);
	
			res = (axp * V(i + 1, j,     k    ) + ...
		       	axm * V(i - 1, j,     k    ) + ...
		       	ayp * V(i,     j + 1, k    ) + ...
		       	aym * V(i,     j - 1, k    ) + ...
		       	azp * V(i,     j,     k + 1) + ...
		       	azm * V(i,     j,     k - 1)) ...
				/ (axp + axm + ayp + aym + azp + azm) - V(i, j, k);
			%res = omega * (1 - sign(idVolt(i, j, k))) * res;
			res = omega * (idVolt(i, j, k) == 0) * res;
			%res = omega * res;
			V(i, j, k) = V(i, j, k) + res;
			res2 = res2 + res^2;
		%end
	end
	end
	end
end

% SOR (red-black)
function [V, res2] = update_rb(oe, V, Mesh, idVolt, Epsr, omega)
	res2 = 0;

	% mesh
	nx = Mesh.nx;
	ny = Mesh.ny;
	nz = Mesh.nz;
	rxp = Mesh.rxp;
	rxm = Mesh.rxm;
	ryp = Mesh.ryp;
	rym = Mesh.rym;
	rzp = Mesh.rzp;
	rzm = Mesh.rzm;

	for k = 2 : nz - 1
	for j = 2 : ny - 1
		imin = 3 - (mod(j + k, 2) == oe);
		imax = nx - 1 - mod(nx - 1 - imin, 2);
		for i = imin : 2 : imax
			%if (idVolt(i, j, k) == 0)
				e0 = Epsr(i, j, k);
	
				axp = (e0 + Epsr(i + 1, j,     k    )) * rxp(i);
				axm = (e0 + Epsr(i - 1, j,     k    )) * rxm(i);
				ayp = (e0 + Epsr(i,     j + 1, k    )) * ryp(j);
				aym = (e0 + Epsr(i,     j - 1, k    )) * rym(j);
				azp = (e0 + Epsr(i,     j,     k + 1)) * rzp(k);
				azm = (e0 + Epsr(i,     j,     k - 1)) * rzm(k);
	
				res = ( ...
					axp * V(i + 1, j,     k    ) + ...
					axm * V(i - 1, j,     k    ) + ...
					ayp * V(i,     j + 1, k    ) + ...
					aym * V(i,     j - 1, k    ) + ...
					azp * V(i,     j,     k + 1) + ...
					azm * V(i,     j,     k - 1)) ...
				 / (axp + axm + ayp + aym + azp + azm) - V(i, j, k);
				%res = omega * (1 - sign(idVolt(i, j, k))) * res;
				res = omega * (idVolt(i, j, k) == 0) * res;
				%res = omega * res;
				V(i, j, k) = V(i, j, k) + res;
				res2 = res2 + res^2;
			%end
		end
	end
	end
end

% boundary condition
function V = bc(V, idVolt, nx, ny, nz)
%
	% Neumann

	% X face
	for k = 1 : nz
	for j = 1 : ny
	%for k = 2 : nz - 1
	%for j = 2 : ny - 1
		if (idVolt(1, j, k) == 0)
			V(1, j, k) = V(2, j, k);
		end
		if (idVolt(nx, j, k) == 0)
			V(nx, j, k) = V(nx - 1, j, k);
		end
	end
	end

	% Y face : slow ?
	for k = 1 : nz
	for i = 1 : nx
	%for k = 2 : nz - 1
	%for i = 2 : nx - 1
		if (idVolt(i, 1, k) == 0)
			V(i, 1, k) = V(i, 2, k);
		end
		if (idVolt(i, ny, k) == 0)
			V(i, ny, k) = V(i, ny - 1, k);
		end
	end
	end

	% Z face : slow ??
	for j = 1 : ny
	for i = 1 : nx
	%for j = 2 : ny - 1
	%for i = 2 : nx - 1
		if (idVolt(i, j, 1) == 0)
			V(i, j, 1) = V(i, j, 2);
		end
		if (idVolt(i, j, nz) == 0)
			V(i, j, nz) = V(i, j, nz - 1);
		end
	end
	end
%
%{
	% 1-st order

	% X face
	for k = 1 : nz
	for j = 1 : ny
	%for k = 2 : nz - 1
	%for j = 2 : ny - 1
		if (idVolt(1, j, k) == 0)
			V(1, j, k) = 2 * V(2, j, k) - V(3, j, k);
		end
		if (idVolt(nx, j, k) == 0)
			V(nx, j, k) = 2 * V(nx - 1, j, k) - V(nx - 2, j, k);
		end
	end
	end

	% Y face ?
	for k = 1 : nz
	for i = 1 : nx
	%for k = 2 : nz - 1
	%for i = 2 : nx - 1
		if (idVolt(i, 1, k) == 0)
			V(i, 1, k) = 2 * V(i, 2, k) - V(i, 3, k);
		end
		if (idVolt(i, ny, k) == 0)
			V(i, ny, k) = 2 * V(i, ny - 1, k) - V(i, ny - 2, k);
		end
	end
	end

	% Z face ??
	for j = 1 : ny
	for i = 1 : nx
	%for j = 2 : ny - 1
	%for i = 2 : nx - 1
		if (idVolt(i, j, 1) == 0)
			V(i, j, 1) = 2 * V(i, j, 2) - V(i, j, 3);
		end
		if (idVolt(i, j, nz) == 0)
			V(i, j, nz) = 2 * V(i, j, nz - 1) - V(i, j, nz - 2);
		end
	end
	end
%}
end

% initial value and scaling
function [V, vscale] = initial(Parm, Geom, idVolt, nx, ny, nz)
	V = zeros(nx, ny, nz);

	% vmin, vmax
	vscale(1) = +inf;
	vscale(2) = -inf;
	for n = 1 : length(Geom)
		if (Geom(n).type == 1)
			v = Parm.volt(Geom(n).pid);
			vscale(1) = min(vscale(1), v);
			vscale(2) = max(vscale(2), v);
		end
	end
	if (vscale(1) == vscale(2))
		disp("*** No voltage difference.")
		return;
	end

	% initial value and scaling to (-1,+1)
	v0 = (vscale(1) + vscale(2)) / 2;
	f = 2 / (vscale(2) - vscale(1));
	for k = 1 : nz
	for j = 1 : ny
		li = (idVolt(:, j, k) > 0);  % logical index
		V(li, j, k) = f * (Parm.volt(idVolt(li, j, k)) - v0);
		%V(li, j, k) = f * (V(li, j, k) - v0);
	end
	end
end

% final : rescaling
function V = final(V, vscale, ny, nz)
	v0 = (vscale(1) + vscale(2)) / 2;
	f = (vscale(2) - vscale(1)) / 2;
	for k = 1 : nz
	for j = 1 : ny
		V(:, j, k) = v0 + (f * V(:, j, k));
	end
	end
end
