/*
OpenSTF Version 3.1.0 (CPU + MPI)
*/

#define MAIN
#include "ost.h"
#include "ost_mpi.h"
#undef MAIN
#include "ost_prototype.h"

static void args(int, char *[], int *, int *, int *, int *, int *, char []);

int main(int argc, char *argv[])
{
	const char prog[] = "(CPU+MPI)";
	const char errfmt[] = "*** file %s open error.\n";
	char str[BUFSIZ];
	int ierr = 0;
	double cpu[] = {0, 0, 0};
	FILE *fp_in = NULL, *fp_out = NULL, *fp_log = NULL;

	// initialize MPI
	mpi_init(argc, argv);
	const int io = !commRank;

	// arguments
	runMode = 0;
	int nthread = 1;
	int vector = 0;
	int ev = 0;
	int geom = 0;
	int prompt = 0;
	char fn_in[BUFSIZ] = "";
	args(argc, argv, &nthread, &vector, &ev, &geom, &prompt, fn_in);

	// set number of threads
#ifdef _OPENMP
	omp_set_num_threads(nthread);
#endif

	// input
	if (io) {
		if ((fp_in = fopen(fn_in, "r")) == NULL) {
			printf(errfmt, fn_in);
			fflush(stdout);
			ierr = 1;
		}
		if (!ierr) {
			ierr = input(fp_in, (!geom && (runMode < 2)));
			fclose(fp_in);
		}
	}
	comm_check(ierr, 0, prompt);

	// === solver ===

	if ((runMode == 0) || (runMode == 1)) {
		// cpu time
		const double t0 = comm_cputime();

		// logo
		if (io) {
			sprintf(str, "<<< %s %s Ver.%d.%d.%d >>>", PROGRAM, prog, VERSION_MAJOR, VERSION_MINOR, VERSION_BUILD);
		}

		if (!geom) {
			// open log file
			if (io) {
				if ((fp_log = fopen(FN_log, "w")) == NULL) {
					printf(errfmt, FN_log);
					fflush(stdout);
					ierr = 1;
				}
			}
			comm_check(ierr, 0, prompt);

			// monitor
			if (io) {
				monitor1(fp_log, str);
				sprintf(str, "Process = %d, Thread = %d (%s)", commSize, nthread, (vector ? "vector" : "novector"));
				monitor1(fp_log, str);
			}
		}
		else {
			if (io) {
				printf("%s\n", str);
				fflush(stdout);
			}
		}

		// plot geometry 3d
		if (io) {
			plot3dGeom(ev);
		}

		// plot geometry only : exit
		if (geom) {
			mpi_close();
			exit(0);
		}

		// broadcast (MPI)
		if (commSize > 1) {
			comm_broadcast();
		}

		// setup
		setupsize();
		setupfactor();
		setupid(iMin2, iMax2, jMin2, jMax2, kMin2, kMax2);

		// monitor
		if (io) {
			monitor2(fp_log, 0, vector, commSize);
		}

		// solve
		const double t1 = comm_cputime();
		sor(vector, fp_log);
		const double t2 = comm_cputime();

		// energy and charge
		echar(iMin, iMax, jMin, jMax, kMin, kMax, iMin2, jMin2, kMin2);
		if (commSize > 1) {
			comm_reduction();
		}

		// gather
		if ((runMode == 1) && (commSize > 1)) {
			comm_gather();
		}

		if (io) {
			// monitor
			monitor3(fp_log);

			// output files
			monitor4(fp_log, ev);

			// write ost.out
			if (runMode == 1) {
				if ((fp_out = fopen(FN_out, "wb")) == NULL) {
					printf(errfmt, FN_out);
					getchar();
					exit(1);
				}
				writeout(fp_out);
				fclose(fp_out);
			}
		}

		// cpu time
		const double t3 = comm_cputime();
		cpu[0] = t1 - t0;
		cpu[1] = t2 - t1;
		cpu[2] = t3 - t2;

		if (io) {
			// cpu time
			monitor5(fp_log, cpu);

			// close log file
			fclose(fp_log);
		}
	}

	// === post ===

	if ((runMode == 0) || (io && (runMode == 2))) {
		// read ost.out
		if (io && (runMode == 2)) {
			if ((fp_out = fopen(FN_out, "rb")) == NULL) {
				printf(errfmt, FN_out);
				getchar();
				exit(1);
			}
			readout(fp_out);
			fclose(fp_out);
		}

		// make data
		if (NL1d > 0) {
			if      (runMode == 0) {
				makeL1d(iMin, iMax, jMin, jMax, kMin, kMax, iMin2, jMin2, kMin2);
				if (commSize > 1) {
					comm_gather1d();
				}
			}
			else if (runMode == 2) {
				makeL1d(0, Nx, 0, Ny, 0, Nz, 0, 0, 0);
			}
		}
		if (NP2d > 0) {
			if      (runMode == 0) {
				makeP2d(iMin, iMax, jMin, jMax, kMin, kMax, iMin2, jMin2, kMin2);
				if (commSize > 1) {
					comm_gather2d();
				}
			}
			else if (runMode == 2) {
				makeP2d(0, Nx, 0, Ny, 0, Nz, 0, 0, 0);
			}
		}

		// post process
		if (io) {
			post(ev);
		}
	}

	// finalize MPI
	mpi_close();

	// prompt
	if (io && prompt) getchar();

	return 0;
}

static void args(int argc, char *argv[],
	int *nthread, int *vector, int *ev, int *geom, int *prompt, char fn_in[])
{
	const char usage[] = "Usage : mpiexec -n <process> ost_mpi [-solver|-post] [-n <thread>] [-vector|-novector] [-ev] [-geom] [-prompt] <datafile>";

	if (argc < 2) {
		printf("%s\n", usage);
		exit(0);
	}

	while (--argc) {
		++argv;
		if      (!strcmp(*argv, "-solver")) {
			runMode = 1;
		}
		else if (!strcmp(*argv, "-post")) {
			runMode = 2;
		}
		else if (!strcmp(*argv, "-n")) {
			if (--argc) {
				*nthread = atoi(*++argv);
				if (*nthread <= 0) *nthread = 1;
			}
			else {
				break;
			}
		}
		else if (!strcmp(*argv, "-vector")) {
			*vector = 1;
		}
		else if (!strcmp(*argv, "-novector")) {
			*vector = 0;
		}
		else if (!strcmp(*argv, "-ev")) {
			*ev = 1;
		}
		else if (!strcmp(*argv, "-geom")) {
			*geom = 1;
		}
		else if (!strcmp(*argv, "-prompt")) {
			*prompt = 1;
		}
		else if (!strcmp(*argv, "--help")) {
			if (commRank == 0) {
				printf("%s\n", usage);
				fflush(stdout);
			}
			mpi_close();
			exit(0);
		}
		else {
			strcpy(fn_in, *argv);
		}
	}
}
