/*
comm.c (MPI)

MPI routines
*/

#ifdef _MPI
#include <mpi.h>
#endif

#include "ost.h"
#include "ost_mpi.h"

// initialize
void mpi_init(int argc, char **argv)
{
#ifdef _MPI
	MPI_Init(&argc, &argv);
	MPI_Comm_size(MPI_COMM_WORLD, &commSize);
	MPI_Comm_rank(MPI_COMM_WORLD, &commRank);
#else
	commSize = 1;
	commRank = 0;
	argc = argc;	// dummy
	argv = argv;	// dummy
#endif
}

// close
void mpi_close(void)
{
#ifdef _MPI
	MPI_Finalize();
#endif
}

// check error code
// mode = 0/1 : Bcast/Allreduce
void comm_check(int ierr, int mode, int prompt)
{
#ifdef _MPI
	if (commSize > 1) {
		if (mode == 0) {
			MPI_Bcast(&ierr, 1, MPI_INT, 0, MPI_COMM_WORLD);
		}
		else {
			int g_ierr;
			MPI_Allreduce(&ierr, &g_ierr, 1, MPI_INT, MPI_LOR, MPI_COMM_WORLD);
			ierr = g_ierr;
		}
	}
	if (ierr) {
		MPI_Finalize();
	}
#endif
	mode = mode;  // dummy
	if (ierr) {
		if (prompt && (commRank == 0)) {
			getchar();
		}
		exit(0);
	}
}

// get cpu time [sec]
double comm_cputime(void)
{
#ifdef _MPI
	MPI_Barrier(MPI_COMM_WORLD);
	return MPI_Wtime();
#else
#ifdef _WIN32
	return (double)clock() / CLOCKS_PER_SEC;
#else
	struct timespec ts;
	clock_gettime(CLOCK_REALTIME, &ts);
	return (ts.tv_sec + (ts.tv_nsec * 1e-9));
#endif  // _WIN32
#endif  // MPI
}

// gather string
void comm_string(const char *str, char *lstr)
{
#ifdef _MPI
	char buff[BUFSIZ];
	if (commRank == 0) {
		MPI_Status status;
		strcpy(lstr, str);
		for (int i = 1; i < commSize; i++) {
			MPI_Recv(buff, BUFSIZ, MPI_CHAR, i, 0, MPI_COMM_WORLD, &status);
			strcat(lstr, "\n");
			strcat(lstr, buff);
		}
	}
	else {
		strcpy(buff, str);
		MPI_Send(buff, BUFSIZ, MPI_CHAR, 0, 0, MPI_COMM_WORLD);
	}
#else
	strcpy(lstr, str);
#endif
}

// broadcast input data
void comm_broadcast(void)
{
#ifdef _MPI
	const int maxvolt = sizeof(Volt) / sizeof(double);  // =256
	const int maxepsr = sizeof(Epsr) / sizeof(real_t);  // =256

	int    *i_buf = NULL;
	double *d_buf = NULL;
	char   *c_buf = NULL;
	int i_num = 0;
	int d_num = 0;
	int c_num = 0;

	// variables to buffers (root only)

	if (commRank == 0) {
		// number of data
		i_num = 10 + (3 * NGeom) + (2 * NL1d) + (1 * NP2d);
		d_num = 2 + (Nx + 1) + (Ny + 1) + (Nz + 1) + (1 * maxvolt) + (1 * maxepsr) + (6 * NGeom);
		c_num = (1 * NL1d) + (1 * NP2d);

		// alloc
		i_buf =    (int *)malloc(i_num * sizeof(int));
		d_buf = (double *)malloc(d_num * sizeof(double));
		c_buf =   (char *)malloc(c_num * sizeof(char));

		int i_id = 0;
		int d_id = 0;
		int c_id = 0;

		i_buf[i_id++] = Nx;
		i_buf[i_id++] = Ny;
		i_buf[i_id++] = Nz;
		i_buf[i_id++] = NVolt;
		i_buf[i_id++] = NEpsr;
		i_buf[i_id++] = NGeom;
		i_buf[i_id++] = Solver.maxiter;
		i_buf[i_id++] = Solver.nout;
		d_buf[d_id++] = Solver.omega;
		d_buf[d_id++] = Solver.converg;

		for (int i = 0; i <= Nx; i++) {
			d_buf[d_id++] = Xn[i];
		}

		for (int j = 0; j <= Ny; j++) {
			d_buf[d_id++] = Yn[j];
		}

		for (int k = 0; k <= Nz; k++) {
			d_buf[d_id++] = Zn[k];
		}

		for (int n = 0; n < maxvolt; n++) {
			d_buf[d_id++] = Volt[n];
		}

		for (int n = 0; n < maxepsr; n++) {
			d_buf[d_id++] = Epsr[n];
		}

		for (int n = 0; n < NGeom; n++) {
			i_buf[i_id++] = Geom[n].type;
			i_buf[i_id++] = Geom[n].shape;
			i_buf[i_id++] = Geom[n].pid;
			for (int m = 0; m < 6; m++) {
				d_buf[d_id++] = Geom[n].g[m];
			}
		}

		i_buf[i_id++] = NL1d;
		for (int n = 0; n < NL1d; n++) {
			c_buf[c_id++] = L1d[n].direction;
			i_buf[i_id++] = L1d[n].posid[0];
			i_buf[i_id++] = L1d[n].posid[1];
		}

		i_buf[i_id++] = NP2d;
		for (int n = 0; n < NP2d; n++) {
			c_buf[c_id++] = P2d[n].direction;
			i_buf[i_id++] = P2d[n].posid;
		}

		// check
		assert(i_id == i_num);
		assert(d_id == d_num);
		assert(c_id == c_num);
	}

	// broadcast (root to non-root)

	MPI_Bcast(&i_num, 1, MPI_INT, 0, MPI_COMM_WORLD);
	MPI_Bcast(&d_num, 1, MPI_INT, 0, MPI_COMM_WORLD);
	MPI_Bcast(&c_num, 1, MPI_INT, 0, MPI_COMM_WORLD);

	// alloc
	if (commRank > 0) {
		i_buf =    (int *)malloc(i_num * sizeof(int));
		d_buf = (double *)malloc(d_num * sizeof(double));
		c_buf =   (char *)malloc(c_num * sizeof(char));
	}

	MPI_Bcast(i_buf, i_num, MPI_INT,    0, MPI_COMM_WORLD);
	MPI_Bcast(d_buf, d_num, MPI_DOUBLE, 0, MPI_COMM_WORLD);
	MPI_Bcast(c_buf, c_num, MPI_CHAR,   0, MPI_COMM_WORLD);

	if (commRank > 0) {
		int i_id = 0;
		int d_id = 0;
		int c_id = 0;

		Nx             = i_buf[i_id++];
		Ny             = i_buf[i_id++];
		Nz             = i_buf[i_id++];
		NVolt          = i_buf[i_id++];
		NEpsr          = i_buf[i_id++];
		NGeom          = i_buf[i_id++];

		Solver.maxiter = i_buf[i_id++];
		Solver.nout    = i_buf[i_id++];
		Solver.omega   = d_buf[d_id++];
		Solver.converg = d_buf[d_id++];

		Xn = (double *)malloc((Nx + 1) * sizeof(double));
		for (int i = 0; i <= Nx; i++) {
			Xn[i] = d_buf[d_id++];
		}

		Yn = (double *)malloc((Ny + 1) * sizeof(double));
		for (int j = 0; j <= Ny; j++) {
			Yn[j] = d_buf[d_id++];
		}

		Zn = (double *)malloc((Nz + 1) * sizeof(double));
		for (int k = 0; k <= Nz; k++) {
			Zn[k] = d_buf[d_id++];
		}

		for (int n = 0; n < maxvolt; n++) {
			Volt[n] = d_buf[d_id++];
		}

		for (int n = 0; n < maxepsr; n++) {
			Epsr[n] = (real_t)d_buf[d_id++];
		}

		Geom = (geometry_t *)malloc(NGeom * sizeof(geometry_t));
		for (int n = 0; n < NGeom; n++) {
			Geom[n].type  =       i_buf[i_id++];
			Geom[n].shape =       i_buf[i_id++];
			Geom[n].pid   = (id_t)i_buf[i_id++];
			for (int m = 0; m < 6; m++) {
				Geom[n].g[m] = d_buf[d_id++];
			}
		}

		NL1d = i_buf[i_id++];
		L1d = (l1d_t *)malloc(NL1d * sizeof(l1d_t));
		for (int n = 0; n < NL1d; n++) {
			L1d[n].direction = c_buf[c_id++];
			L1d[n].posid[0]  = i_buf[i_id++];
			L1d[n].posid[1]  = i_buf[i_id++];
		}

		NP2d = i_buf[i_id++];
		P2d = (p2d_t *)malloc(NP2d * sizeof(p2d_t));
		for (int n = 0; n < NP2d; n++) {
			P2d[n].direction = c_buf[c_id++];
			P2d[n].posid     = i_buf[i_id++];
		}

		// check
		assert(i_id == i_num);
		assert(d_id == d_num);
		assert(c_id == c_num);
	}

	// free
	free(i_buf);
	free(d_buf);
	free(c_buf);
/*
	// debug
	printf("%d %d\n", commSize, commRank);
	printf("%d %d %d\n", Nx, Ny, Nz);
	//printf("%d %d %g %g\n", Solver.maxiter, Solver.nout, Solver.omega, Solver.converg);
	//for (int i = 0; i <= Nx; i++) printf("%d %f\n", i, Xn[i]);
	//for (int j = 0; j <= Ny; j++) printf("%d %f\n", j, Yn[j]);
	//for (int k = 0; k <= Nz; k++) printf("%d %f\n", k, Zn[k]);
	//printf("%d %d\n", NVolt, NEpsr);
	//for (int n = 0; n < NVolt; n++) printf("%d %f\n", n, Volt[n]);
	//for (int n = 0; n < NEpsr; n++) printf("%d %f\n", n, Epsr[n]);
	//printf("%d\n", NGeom);
	//for (int n = 0; n < NGeom; n++) printf("%d %d %d %d %f %f %f %f %f %f\n", n, Geom[n].type, Geom[n].shape, Geom[n].pid, Geom[n].g[0], Geom[n].g[1], Geom[n].g[2], Geom[n].g[3], Geom[n].g[4], Geom[n].g[5]);
	fflush(stdout);
*/
#endif
}

// sum
real_t comm_sum(real_t l_sum)
{
	real_t sum;
#ifdef _MPI
	MPI_Allreduce(&l_sum, &sum, 1, MPI_REAL_T, MPI_SUM, MPI_COMM_WORLD);
#else
	sum = l_sum;
#endif
	return sum;
}

// reduction
void comm_reduction(void)
{
	for (int n = 0; n < NVolt; n++) {
		Echar[n] = comm_sum((real_t)Echar[n]);
	}
}

// share boundary
void comm_boundary(real_t *sendbuf, real_t *recvbuf)
{
#ifdef _MPI
	MPI_Status status;
	const int tag = 0;
	const int count = (Ny + 1) * (Nz + 1);

	// -X boundary
	if (commRank > 0) {
		// copy to buffer
		int isend = 0;
		for (int j = 0; j <= Ny; j++) {
		for (int k = 0; k <= Nz; k++) {
			sendbuf[isend++] = V[D(1, j, k)];
		}
		}
		assert(isend == count);

		// MPI
		int dst = commRank - 1;
		MPI_Sendrecv(sendbuf, count, MPI_REAL_T, dst, tag,
		             recvbuf, count, MPI_REAL_T, dst, tag, MPI_COMM_WORLD, &status);

		// copy from buffer
		int irecv = 0;
		for (int j = 0; j <= Ny; j++) {
		for (int k = 0; k <= Nz; k++) {
			V[D(0, j, k)] = recvbuf[irecv++];
		}
		}
		assert(irecv == count);
	}

	// +X boundary
	if (commRank < commSize - 1) {
		// copy to buffer
		int isend = 0;
		for (int j = 0; j <= Ny; j++) {
		for (int k = 0; k <= Nz; k++) {
			sendbuf[isend++] = V[D(iMax2 - iMin2 - 1, j, k)];
		}
		}
		assert(isend == count);

		// MPI
		int dst = commRank + 1;
		MPI_Sendrecv(sendbuf, count, MPI_REAL_T, dst, tag,
		             recvbuf, count, MPI_REAL_T, dst, tag, MPI_COMM_WORLD, &status);

		// copy from buffer
		int irecv = 0;
		for (int j = 0; j <= Ny; j++) {
		for (int k = 0; k <= Nz; k++) {
			V[D(iMax2 - iMin2, j, k)] = recvbuf[irecv++];
		}
		}
		assert(irecv == count);
	}
#endif
}

// gather
void comm_gather(void)
{
#ifdef _MPI
	MPI_Status status;
	const int tag = 0;

	if (commRank == 0) {
		// realloc
		V      = (real_t *)realloc(V,      (Nx + 3) * (Ny + 3) * (Nz + 3) * sizeof(real_t));
		idVolt =   (id_t *)realloc(idVolt, (Nx + 3) * (Ny + 3) * (Nz + 3) * sizeof(id_t));
		idEpsr =   (id_t *)realloc(idEpsr, (Nx + 3) * (Ny + 3) * (Nz + 3) * sizeof(id_t));
		// recv
		for (int rank = 1; rank < commSize; rank++) {
			int imin, imax;
			MPI_Recv(&imin, 1, MPI_INT, rank, tag, MPI_COMM_WORLD, &status);
			MPI_Recv(&imax, 1, MPI_INT, rank, tag, MPI_COMM_WORLD, &status);
			//printf("%d %d %d\n", rank, imin, imax); fflush(stdout);
			int count = (imax - imin + 1) * (Ny + 3) * (Nz + 3);
			MPI_Recv(     &V[D(imin, -1, -1)], count, MPI_REAL_T, rank, tag, MPI_COMM_WORLD, &status);
			MPI_Recv(&idVolt[D(imin, -1, -1)], count, MPI_ID_T,   rank, tag, MPI_COMM_WORLD, &status);
			MPI_Recv(&idEpsr[D(imin, -1, -1)], count, MPI_ID_T,   rank, tag, MPI_COMM_WORLD, &status);
		}
	}
	else {
		// send
		int count = (iMax - iMin + 1) * (Ny + 3) * (Nz + 3);
		MPI_Send(&iMin, 1, MPI_INT, 0, tag, MPI_COMM_WORLD);
		MPI_Send(&iMax, 1, MPI_INT, 0, tag, MPI_COMM_WORLD);
		MPI_Send(     &V[D(1, -1, -1)], count, MPI_REAL_T, 0, tag, MPI_COMM_WORLD);
		MPI_Send(&idVolt[D(1, -1, -1)], count, MPI_ID_T,   0, tag, MPI_COMM_WORLD);
		MPI_Send(&idEpsr[D(1, -1, -1)], count, MPI_ID_T,   0, tag, MPI_COMM_WORLD);
	}
#endif
}

// gather 1d
void comm_gather1d(void)
{
#ifdef _MPI
	MPI_Status status;
	const int ncompo = 10;
	const int length = MAX(Nx + 1, MAX(Ny + 1, Nz + 1));
	const int count = NL1d * ncompo * length;
	const int tag = 0;

	if (commRank == 0) {
		double *recvbuf = (double *)malloc(count * sizeof(double));
		for (int rank = 1; rank < commSize; rank++) {
			MPI_Recv(recvbuf, count, MPI_DOUBLE, rank, tag, MPI_COMM_WORLD, &status);
			int id = 0;
			for (int n = 0; n < NL1d; n++) {
				for (int ic = 0; ic < ncompo; ic++) {
					for (int l = 0; l < length; l++) {
						F1d[n][ic][l] += recvbuf[id++];
					}
				}
			}
		}
	}
	else {
		double *sendbuf = (double *)malloc(count * sizeof(double));
		int id = 0;
		for (int n = 0; n < NL1d; n++) {
			for (int ic = 0; ic < ncompo; ic++) {
				for (int l = 0; l < length; l++) {
					sendbuf[id++] = F1d[n][ic][l];
				}
			}
		}
		MPI_Send(sendbuf, count, MPI_DOUBLE, 0, tag, MPI_COMM_WORLD);
	}
#endif
}

// gather 2d
void comm_gather2d(void)
{
#ifdef _MPI
	MPI_Status status;
	const int ncompo = 10;
	const int length = MAX(Nx + 1, MAX(Ny + 1, Nz + 1));
	const int count = NP2d * ncompo * length * length;
	const int tag = 0;

	if (commRank == 0) {
		double *recvbuf = (double *)malloc(count * sizeof(double));
		for (int rank = 1; rank < commSize; rank++) {
			MPI_Recv(recvbuf, count, MPI_DOUBLE, rank, tag, MPI_COMM_WORLD, &status);
			int id = 0;
			for (int n = 0; n < NP2d; n++) {
				for (int ic = 0; ic < ncompo; ic++) {
					for (int l1 = 0; l1 < length; l1++) {
						for (int l2 = 0; l2 < length; l2++) {
							F2d[n][ic][l1][l2] += recvbuf[id++];
						}
					}
				}
			}
		}
	}
	else {
		double *sendbuf = (double *)malloc(count * sizeof(double));
		int id = 0;
		for (int n = 0; n < NP2d; n++) {
			for (int ic = 0; ic < ncompo; ic++) {
				for (int l1 = 0; l1 < length; l1++) {
					for (int l2 = 0; l2 < length; l2++) {
						sendbuf[id++] = F2d[n][ic][l1][l2];
					}
				}
			}
		}
		MPI_Send(sendbuf, count, MPI_DOUBLE, 0, tag, MPI_COMM_WORLD);
	}
#endif
}
