/*
sor.c (MPI)

SOR method
*/

#include "ost.h"
#include "ost_mpi.h"
#include "ost_prototype.h"

// red-black (novector)
static real_t update_rb_novector(int oe)
{
	real_t res2 = 0;

	const real_t omega = (real_t)Solver.omega;

	const int64_t ni = D(1, 0, 0) - D(0, 0, 0);
	const int64_t nj = D(0, 1, 0) - D(0, 0, 0);
	const int64_t nk = D(0, 0, 1) - D(0, 0, 0);

	int64_t i;
#ifdef _OPENMP
#pragma omp parallel for reduction (+: res2)
#endif
	for (        i = iMin; i <= iMax; i++) {
	for (int64_t j = jMin; j <= jMax; j++) {
		int64_t kmin = kMin + (((i + j) % 2 == oe) ? 0 : 1);
		int64_t kmax = kMax;
		for (int64_t k = kmin; k <= kmax; k += 2) {
			const int64_t n = D(i - iMin2, j - jMin2, k - kMin2);
			const real_t e = Epsr[idEpsr[n]];

			const real_t axp = (e + Epsr[idEpsr[n + ni]]) * RXp[i];
			const real_t axm = (e + Epsr[idEpsr[n - ni]]) * RXm[i];
			const real_t ayp = (e + Epsr[idEpsr[n + nj]]) * RYp[j];
			const real_t aym = (e + Epsr[idEpsr[n - nj]]) * RYm[j];
			const real_t azp = (e + Epsr[idEpsr[n + nk]]) * RZp[k];
			const real_t azm = (e + Epsr[idEpsr[n - nk]]) * RZm[k];

			const real_t res = (idVolt[n] ? 0 : 1) * omega * ((
				axp * V[n + ni] +
				axm * V[n - ni] +
				ayp * V[n + nj] +
				aym * V[n - nj] +
				azp * V[n + nk] +
				azm * V[n - nk]) / (axp + axm + ayp + aym + azp + azm) - V[n]);
			V[n] += res;
			res2 += res * res;
		}
	}
	}

	return res2;
}

// red-black (vector)
static real_t update_rb_vector(int oe)
{
	real_t res2 = 0;

	const real_t omega = (real_t)Solver.omega;

	const int64_t ni = D(1, 0, 0) - D(0, 0, 0);
	const int64_t nj = D(0, 1, 0) - D(0, 0, 0);
	const int64_t nk = D(0, 0, 1) - D(0, 0, 0);

	int64_t i;
#ifdef _OPENMP
#pragma omp parallel for reduction (+: res2)
#endif
	for (        i = iMin; i <= iMax; i++) {
	for (int64_t j = jMin; j <= jMax; j++) {
		int64_t kmin = kMin + ((i + j) % 2 == oe) ? 0 : 1;
		int64_t kmax = kMax;
		const int64_t n = D(i - iMin2, j - jMin2, kmin - kMin2);
		real_t *ptr = &V[n];
		real_t *ptr_vxp = &V[n + ni];
		real_t *ptr_vxm = &V[n - ni];
		real_t *ptr_vyp = &V[n + nj];
		real_t *ptr_vym = &V[n - nj];
		real_t *ptr_vzp = &V[n + nk];
		real_t *ptr_vzm = &V[n - nk];
		real_t *ptr_e   = &Epsr_v[n];
		real_t *ptr_exp = &Epsr_v[n + ni];
		real_t *ptr_exm = &Epsr_v[n - ni];
		real_t *ptr_eyp = &Epsr_v[n + nj];
		real_t *ptr_eym = &Epsr_v[n - nj];
		real_t *ptr_ezp = &Epsr_v[n + nk];
		real_t *ptr_ezm = &Epsr_v[n - nk];
		real_t rxp = RXp[i];
		real_t rxm = RXm[i];
		real_t ryp = RYp[j];
		real_t rym = RYm[j];
		real_t *ptr_rzp = &RZp[kmin];
		real_t *ptr_rzm = &RZm[kmin];
		int    *ptr_id = &idVolt_v[n];
#ifdef __NEC__
#pragma _NEC ivdep
#elif __FUJITSU
#pragma loop novrec
#endif
		for (int64_t k = kmin; k <= kmax; k += 2) {
			const real_t e = *ptr_e;

			const real_t axp = (e + (*ptr_exp)) * rxp;
			const real_t axm = (e + (*ptr_exm)) * rxm;
			const real_t ayp = (e + (*ptr_eyp)) * ryp;
			const real_t aym = (e + (*ptr_eym)) * rym;
			const real_t azp = (e + (*ptr_ezp)) * (*ptr_rzp);
			const real_t azm = (e + (*ptr_ezm)) * (*ptr_rzm);

			const real_t res = (*ptr_id ? 0 : 1) * omega * ((
				axp * (*ptr_vxp) +
				axm * (*ptr_vxm) +
				ayp * (*ptr_vyp) +
				aym * (*ptr_vym) +
				azp * (*ptr_vzp) +
				azm * (*ptr_vzm)) / (axp + axm + ayp + aym + azp + azm) - (*ptr));

			*ptr += res;
			res2 += res * res;

			ptr += 2;
			ptr_vxp += 2;
			ptr_vxm += 2;
			ptr_vyp += 2;
			ptr_vym += 2;
			ptr_vzp += 2;
			ptr_vzm += 2;
			ptr_e += 2;
			ptr_exp += 2;
			ptr_exm += 2;
			ptr_eyp += 2;
			ptr_eym += 2;
			ptr_ezp += 2;
			ptr_ezm += 2;
			ptr_id += 2;
			ptr_rzp += 2;
			ptr_rzm += 2;
		}
	}
	}

	return res2;
}

void sor(int vector, FILE *fp)
{
	char str[BUFSIZ];
	const char msg1[] = "  --- converged ---";
	const char msg2[] = "  --- diverged ---";
	const char msg3[] = "  --- not converged ---";
	const char fmt[] = "   %7d    %.7f";

	int64_t num = (iMax2 - iMin2 + 3) * (jMax2 - jMin2 + 3) * (kMax2 - kMin2 + 3);

	// V
	V = (real_t *)malloc(num * sizeof(real_t));
	memset(V, 0, num * sizeof(real_t));

	// vector
	if (vector) {
		// Er array
		Epsr_v = (real_t *)malloc(num * sizeof(real_t));
		for (int64_t n = 0; n < num; n++) {
			Epsr_v[n] = Epsr[idEpsr[n]];  // direct access
		}
		// Volt array
		idVolt_v = (int *)malloc(num * sizeof(int));
		for (int64_t n = 0; n < num; n++) {
			idVolt_v[n] = idVolt[n];  // id_t -> int
		}
	}

	// residual
	const size_t r_size = (Solver.maxiter + 1) * sizeof(double);
	Residual = (double *)malloc(r_size);
	memset(Residual, 0, r_size);

	// MPI buffer
	const size_t b_size = (Ny + 1) * (Nz + 1) * sizeof(real_t);
	real_t *sendbuf = (real_t *)malloc(b_size);
	real_t *recvbuf = (real_t *)malloc(b_size);

	// electrode
	electrode(iMin2, iMax2, jMin2, jMax2, kMin2, kMax2);

	// initial field
	real_t vmin, vmax;
	initial(&vmin, &vmax, iMin2, iMax2, jMin2, jMax2, kMin2, kMax2);

	// iteration
	int iter;
	for (iter = 0; iter <= Solver.maxiter; iter++) {
		// update
		real_t res2 = 0;
		if ((vector == 0) || (vector == 1)) {
			for (int oe = 0; oe < 2; oe++) {
				if (vector == 1) {
					res2 += update_rb_vector(oe);
				}
				else {
					res2 += update_rb_novector(oe);
				}
				if (commSize > 1) {
					comm_boundary(sendbuf, recvbuf);
				}
			}
		}

		// residual
		res2 = comm_sum(res2);
		Residual[iter] = sqrt(res2 / ((double)(Nx + 1) * (Ny + 1) * (Nz + 1)));

		// monitor
		if ((iter % Solver.nout) == 0 || (Residual[iter] < Solver.converg)) {
			sprintf(str, fmt, iter, Residual[iter]);
			if (commRank == 0) monitor1(fp, str);
		}

		// converged / diverged / not converged
		if (Residual[iter] < Solver.converg) {
			if (commRank == 0) monitor1(fp, msg1);
			break;
		}
		if (Residual[iter] > 1e10) {
			if (commRank == 0) monitor1(fp, msg2);
			break;
		}
		if (iter == Solver.maxiter) {
			if (commRank == 0) monitor1(fp, msg3);
			break;
		}
	}

	// iterations
	NumIter = iter + 1;

	// final
	final(vmin, vmax, iMin2, iMax2, jMin2, jMax2, kMin2, kMax2);

	// edge
	edge(iMin2, iMax2, jMin2, jMax2, kMin2, kMax2);

	// free
	free(sendbuf);
	free(recvbuf);
	if (vector) {
		if (Epsr_v   != NULL) free(Epsr_v);
		if (idVolt_v != NULL) free(idVolt_v);
	}
}
