/*
OpenSTF Version 3.1.0 (CPU + OpenMP)
*/

#define MAIN
#include "ost.h"
#undef MAIN
#include "ost_prototype.h"

static void args(int, char *[], int *, int *, int *, int *, int *, char []);

int main(int argc, char *argv[])
{
	const char prog[] = "(CPU+OpenMP)";
	const char errfmt[] = "*** file %s open error.\n";
	char str[BUFSIZ];
	int ierr = 0;
	double cpu[] = {0, 0, 0};
	FILE *fp_in = NULL, *fp_out = NULL, *fp_log = NULL;

	const int io = 1;

	// arguments
	runMode = 0;
	int nthread = 1;
	int vector = 0;
	int ev = 0;
	int geom = 0;
	int prompt = 0;
	char fn_in[BUFSIZ] = "";
	args(argc, argv, &nthread, &vector, &ev, &geom, &prompt, fn_in);

	// set number of threads
#ifdef _OPENMP
	omp_set_num_threads(nthread);
#endif

	// input
	if (io) {
		if ((fp_in = fopen(fn_in, "r")) == NULL) {
			printf(errfmt, fn_in);
			fflush(stdout);
			ierr = 1;
		}
		if (!ierr) {
			ierr = input(fp_in, (!geom && (runMode < 2)));
			fclose(fp_in);
		}
	}
	if (ierr) {
		if (prompt) getchar();
		exit(1);
	}

	// === solver ===

	if ((runMode == 0) || (runMode == 1)) {
		// cpu time
		const double t0 = cputime();

		// logo
		if (io) {
			sprintf(str, "<<< %s %s Ver.%d.%d.%d >>>", PROGRAM, prog, VERSION_MAJOR, VERSION_MINOR, VERSION_BUILD);
		}

		if (!geom) {
			// open log file
			if (io) {
				if ((fp_log = fopen(FN_log, "w")) == NULL) {
					printf(errfmt, FN_log);
					fflush(stdout);
					ierr = 1;
				}
			}
			if (ierr) {
				if (prompt) getchar();
				exit(1);
			}

			// monitor
			if (io) {
				monitor1(fp_log, str);
				sprintf(str, "Thread = %d (%s)", nthread, (vector == 1 ? "vector" : vector == 0 ? "novector" : vector == -1 ? "serial" : vector == -2 ? "Neumann BC" : ""));
				monitor1(fp_log, str);
			}
		}
		else {
			if (io) {
				printf("%s\n", str);
				fflush(stdout);
			}
		}

		// plot geometry 3d
		if (io) {
			plot3dGeom(ev);
		}

		// plot geometry only : exit
		if (geom) {
			exit(0);
		}

		// setup
		setupfactor();
		setupid(0, Nx, 0, Ny, 0, Nz);

		// monitor
		if (io) {
			monitor2(fp_log, 0, vector, 1);
		}

		// solve
		const double t1 = cputime();
		sor(vector, fp_log);
		const double t2 = cputime();

		// energy and charge
		echar(0, Nx, 0, Ny, 0, Nz, 0, 0, 0);

		if (io) {
			// monitor
			monitor3(fp_log);

			// output files
			monitor4(fp_log, ev);

			// write ost.out
			if (runMode == 1) {
				if ((fp_out = fopen(FN_out, "wb")) == NULL) {
					printf(errfmt, FN_out);
					getchar();
					exit(1);
				}
				writeout(fp_out);
				fclose(fp_out);
			}
		}

		// cpu time
		const double t3 = cputime();
		cpu[0] = t1 - t0;
		cpu[1] = t2 - t1;
		cpu[2] = t3 - t2;

		if (io) {
			// cpu time
			monitor5(fp_log, cpu);

			// close log file
			fclose(fp_log);
		}
	}

	// === post ===

	if ((runMode == 0) || (io && (runMode == 2))) {
		// read ost.out
		if (io && (runMode == 2)) {
			if ((fp_out = fopen(FN_out, "rb")) == NULL) {
				printf(errfmt, FN_out);
				getchar();
				exit(1);
			}
			readout(fp_out);
			fclose(fp_out);
		}

		// make data
		if (NL1d > 0) {
			makeL1d(0, Nx, 0, Ny, 0, Nz, 0, 0, 0);
		}
		if (NP2d > 0) {
			makeP2d(0, Nx, 0, Ny, 0, Nz, 0, 0, 0);
		}

		// post process
		if (io) {
			post(ev);
		}
	}

	// prompt
	if (io && prompt) getchar();

	return 0;
}

static void args(int argc, char *argv[],
	int *nthread, int *vector, int *ev, int *geom, int *prompt, char fn_in[])
{
	const char usage[] = "Usage : ost [-solver|-post] [-n <thread>] [-vector|-novector] [-ev] [-geom] [-prompt] <datafile>";

	if (argc < 2) {
		printf("%s\n", usage);
		exit(0);
	}

	while (--argc) {
		++argv;
		if      (!strcmp(*argv, "-solver")) {
			runMode = 1;
		}
		else if (!strcmp(*argv, "-post")) {
			runMode = 2;
		}
		else if (!strcmp(*argv, "-n")) {
			if (--argc) {
				*nthread = atoi(*++argv);
				if (*nthread <= 0) *nthread = 1;
			}
			else {
				break;
			}
		}
		else if (!strcmp(*argv, "-vector")) {
			*vector = 1;
		}
		else if (!strcmp(*argv, "-novector")) {
			*vector = 0;
		}
		else if (!strcmp(*argv, "-serial")) {
			*vector = -1;
		}
		else if (!strcmp(*argv, "-neumann")) {
			*vector = -2;
		}
		else if (!strcmp(*argv, "-ev")) {
			*ev = 1;
		}
		else if (!strcmp(*argv, "-geom")) {
			*geom = 1;
		}
		else if (!strcmp(*argv, "-prompt")) {
			*prompt = 1;
		}
		else if (!strcmp(*argv, "--help")) {
			printf("%s\n", usage);
			exit(0);
		}
		else {
			strcpy(fn_in, *argv);
		}
	}

	if (*vector == -1) {
		*nthread = 1;
	}
}
