/*
input.c

input data
*/

#include "ost.h"
#include "ost_prototype.h"

#define MAXTOKEN 1000

int input(FILE *fp, int check)
{
	int    ntoken, ngeom, nline;
	int    version = 0;
	int    nxr = 0, nyr = 0, nzr = 0;
	int    *dxr = NULL, *dyr = NULL, *dzr = NULL;
	double *xr = NULL, *yr = NULL, *zr = NULL;
	char   strline[BUFSIZ], strkey[BUFSIZ], strsave[BUFSIZ];
	char   *token[MAXTOKEN];
	const int array_inc = 10000;		// reduce malloc times
	const char sep[] = " \t";			// separator

	// initialize

	NVolt = 1;
	NEpsr = 1;
	Volt[0] = 0;
	Epsr[0] = 1;

	NGeom = 0;
	Geom = NULL;

	Solver.omega = 1.95;
	Solver.maxiter = 1000;
	Solver.nout = 50;
	Solver.converg = 1e-5;

	Piter = 0;

	NL1d = 0;
	L1dScale.db = 0;      // linear
	L1dScale.user = 0;    // auto scale
	L1dLog = 1;

	NP2d = 0;
	P2dFigure[0] = P2dFigure[1] = 1;
	P2dScale.db = 0;      // linear
	P2dScale.user = 0;    // auto scale
	P2dObject[0] = P2dObject[1] = 1;
	P2dZoomin = 0;
	P2dLog = 1;

	Width2d    = 750;
	Height2d   = 500;
	Fontsize2d = 15;
	Fontname2d = 0;

	Width3d    = 600;
	Height3d   = 600;
	Fontsize3d = 12;
	Fontname3d = 0;
	Theta3d    = 60;
	Phi3d      = 30;

	// read

	nline = 0;
	while (fgets(strline, sizeof(strline), fp) != NULL) {
		//printf("%s", strline);

		// skip a vacant line
		if (strlen(strline) <= 1) continue;

		// skip a comment line
		if (strline[0] == '#') continue;

		// delete "\n"
		//printf("%ld\n", strlen(strline));
		if (strstr(strline, "\r\n") != NULL) {
			strline[strlen(strline) - 2] = '\0';
		}
		else if ((strstr(strline, "\r") != NULL) || (strstr(strline, "\n") != NULL)) {
			strline[strlen(strline) - 1] = '\0';
		}
		//printf("%ld\n", strlen(strline));

		// "end" -> break
		if (!strncmp(strline, "end", 3)) break;

		// save "strline"
		strcpy(strsave, strline);

		// token ("strline" is destroyed)
		ntoken = tokenize(strline, sep, token, MAXTOKEN);
		//for (int i = 0; i < ntoken; i++) printf("%d %s\n", i, token[i]);

		// check number of data and "=" (excluding header)
		if ((nline > 0) && ((ntoken < 3) || strcmp(token[1], "="))) continue;

		// keyword
		strcpy(strkey, token[0]);

		// input
		if      (nline == 0) {
			if (strcmp(strkey, "OpenSTF")) {
				fprintf(stderr, "*** not OpenSTF data\n");
				return 1;
			}
			else if (ntoken > 2) {
				version = (10 * atoi(token[1])) + atoi(token[2]);
				version = version;  // dummy
			}
			nline++;
		}
		else if (!strcmp(strkey, "title")) {
			strncpy(Title, strchr(strsave, '=') + 2, sizeof(Title));
		}
		else if (!strcmp(strkey, "xmesh")) {
			if ((ntoken < 5) || (ntoken % 2 == 0)) {
				fprintf(stderr, "*** invalid %s data\n", strkey);
				return 1;
			}
			nxr = (ntoken - 3) / 2;
			xr = (double *)malloc((nxr + 1) * sizeof(double));
			dxr = (int *)malloc(nxr * sizeof(int));
			sscanf(token[2], "%lf", &xr[0]);
			for (int i = 0; i < nxr; i++) {
				sscanf(token[2 * i + 3], "%d", &dxr[i]);
				sscanf(token[2 * i + 4], "%lf", &xr[i + 1]);
			}
		}
		else if (!strcmp(strkey, "ymesh")) {
			if ((ntoken < 5) || (ntoken % 2 == 0)) {
				fprintf(stderr, "*** invalid %s data\n", strkey);
				return 1;
			}
			nyr = (ntoken - 3) / 2;
			yr = (double *)malloc((nyr + 1) * sizeof(double));
			dyr = (int *)malloc(nyr * sizeof(int));
			sscanf(token[2], "%lf", &yr[0]);
			for (int j = 0; j < nyr; j++) {
				sscanf(token[2 * j + 3], "%d", &dyr[j]);
				sscanf(token[2 * j + 4], "%lf", &yr[j + 1]);
			}
		}
		else if (!strcmp(strkey, "zmesh")) {
			if ((ntoken < 5) || (ntoken % 2 == 0)) {
				fprintf(stderr, "*** invalid %s data\n", strkey);
				return 1;
			}
			nzr = (ntoken - 3) / 2;
			zr = (double *)malloc((nzr + 1) * sizeof(double));
			dzr = (int *)malloc(nzr * sizeof(int));
			sscanf(token[2], "%lf", &zr[0]);
			for (int k = 0; k < nzr; k++) {
				sscanf(token[2 * k + 3], "%d", &dzr[k]);
				sscanf(token[2 * k + 4], "%lf", &zr[k + 1]);
			}
		}
		else if (!strcmp(strkey, "volt")) {
			if (NVolt >= 256) {
				fprintf(stderr, "*** too many %s data #%d\n", strkey, NVolt + 1);
				return 1;
			}
			Volt[NVolt++] = atof(token[2]);
		}
		else if (!strcmp(strkey, "epsr")) {
			if (NEpsr >= 256) {
				fprintf(stderr, "*** too many %s data #%d\n", strkey, NEpsr + 1);
				return 1;
			}
			Epsr[NEpsr++] = (real_t)atof(token[2]);
		}
		else if (!strcmp(strkey, "geometry")) {
			if (ntoken < 5) {
				fprintf(stderr, "*** invalid %s data #%d\n", strkey, NGeom + 1);
				return 1;
			}
			if (NGeom % array_inc == 0) {
				Geom = (geometry_t *)realloc(Geom, (NGeom + array_inc) * sizeof(geometry_t));
			}
			Geom[NGeom].type  = atoi(token[2]);        // =1/2
			Geom[NGeom].pid   = (id_t)atoi(token[3]);  // =1,2,...
			Geom[NGeom].shape = atoi(token[4]);        // =1,2,...
			switch (Geom[NGeom].shape) {
				case 1:
				case 2:
				case 11:
				case 12:
				case 13:
					ngeom = 6;
					break;
				case 31:
				case 32:
				case 33:
				case 41:
				case 42:
				case 43:
				case 51:
				case 52:
				case 53:
					ngeom = 8;
					break;
				default:
					ngeom = 0;
					break;
			}
			if (ntoken < 5 + ngeom) {
				fprintf(stderr, "*** invalid %s data #%d\n", strkey, NGeom + 1);
				return 1;
			}
			for (int n = 0; n < ngeom; n++) {
				Geom[NGeom].g[n] = atof(token[5 + n]);
			}
			NGeom++;
		}
		else if (!strcmp(strkey, "solver")) {
			if (ntoken > 5) {
				Solver.omega   = atof(token[2]);
				Solver.maxiter = atoi(token[3]);
				Solver.nout    = atoi(token[4]);
				Solver.converg = atof(token[5]);
			}
		}
		else if (!strcmp(strkey, "plot3dgeom")) {
			// not used
		}

		// === post ===

		else if (!strcmp(strkey, "plotiter")) {
			Piter = atoi(token[2]);
		}
		else if (!strcmp(strkey, "plot1d")) {
			if (ntoken < 6) {
				fprintf(stderr, "*** invalid %s data #%d\n", strkey, NL1d + 1);
				return 1;
			}
			L1d = (l1d_t *)realloc(L1d, (NL1d + 1) * sizeof(l1d_t));
			L1d[NL1d].component = (char)toupper((int)token[2][0]);
			L1d[NL1d].direction = (char)toupper((int)token[3][0]);
			L1d[NL1d].position[0] = atof(token[4]);
			L1d[NL1d].position[1] = atof(token[5]);
			NL1d++;
		}
		else if (!strcmp(strkey, "1ddb")) {
			L1dScale.db = atoi(token[2]);
		}
		else if (!strcmp(strkey, "1dscale")) {
			if (ntoken > 4) {
				L1dScale.user = 1;
				L1dScale.min = atof(token[2]);
				L1dScale.max = atof(token[3]);
				L1dScale.div = atoi(token[4]);
			}
		}
		else if (!strcmp(strkey, "1dlog")) {
			L1dLog = atoi(token[2]);
		}
		else if (!strcmp(strkey, "plot2d")) {
			if (ntoken < 5) {
				fprintf(stderr, "*** invalid %s data #%d\n", strkey, NP2d + 1);
				return 1;
			}
			P2d = (p2d_t *)realloc(P2d, (NP2d + 1) * sizeof(p2d_t));
			strcpy(P2d[NP2d].component, token[2]);
			P2d[NP2d].direction = (char)toupper((int)token[3][0]);
			P2d[NP2d].position = atof(token[4]);
			NP2d++;
		}
		else if (!strcmp(strkey, "2dfigure")) {
			if (ntoken > 3) {
				P2dFigure[0] = atoi(token[2]);
				P2dFigure[1] = atoi(token[3]);
			}
		}
		else if (!strcmp(strkey, "2ddb")) {
			P2dScale.db = atoi(token[2]);
		}
		else if (!strcmp(strkey, "2dscale")) {
			if (ntoken > 3) {
				P2dScale.user = 1;
				P2dScale.min = atof(token[2]);
				P2dScale.max = atof(token[3]);
				// token[4] : nou used
			}
		}
		else if (!strcmp(strkey, "2dcontour")) {
			P2dContour = atoi(token[2]);
		}
		else if (!strcmp(strkey, "2dobject")) {
			if (ntoken > 3) {
				P2dObject[0] = atoi(token[2]);
				P2dObject[1] = atoi(token[3]);
			}
		}
		else if (!strcmp(strkey, "2dzoom")) {
			if (ntoken > 5) {
				P2dZoomin = 1;
				P2dZoom[0][0] = atof(token[2]);
				P2dZoom[0][1] = atof(token[3]);
				P2dZoom[1][0] = atof(token[4]);
				P2dZoom[1][1] = atof(token[5]);
			}
		}
		else if (!strcmp(strkey, "2dlog")) {
			P2dLog = atoi(token[2]);
		}
		else if (!strcmp(strkey, "marker")) {
			// not used
		}
		else if (!strcmp(strkey, "window2d")) {
			if (ntoken > 5) {
				Width2d    = atoi(token[2]);
				Height2d   = atoi(token[3]);
				Fontsize2d = atoi(token[4]);
				Fontname2d = atoi(token[5]);
			}
		}
		else if (!strcmp(strkey, "window3d")) {
			if (version >= 31) {
				if (ntoken > 7) {
					Width3d    = atoi(token[2]);
					Height3d   = atoi(token[3]);
					Fontsize3d = atoi(token[4]);
					Fontname3d = atoi(token[5]);
					Theta3d    = atof(token[6]);
					Phi3d      = atof(token[7]);
				}
			}
			else {
				if (ntoken > 6) {
					Width3d    = atoi(token[2]);
					Height3d   = atoi(token[3]);
					Fontsize3d = atoi(token[4]);
					Fontname3d = 0;
					Theta3d    = atof(token[5]);
					Phi3d      = atof(token[6]);
				}
			}
		}
	}

	// debug
	//printf("title = %s\n", Title);
	//printf("xmesh = %e", xr[0]); for (int i = 0; i < nxr; i++) printf(" %d %e", dxr[i], xr[i + 1]); printf("\n");
	//printf("ymesh = %e", yr[0]); for (int j = 0; j < nyr; j++) printf(" %d %e", dyr[j], yr[j + 1]); printf("\n");
	//printf("zmesh = %e", zr[0]); for (int k = 0; k < nzr; k++) printf(" %d %e", dzr[k], zr[k + 1]); printf("\n");
	//for (int n = 0; n < NVolt; n++) printf("volt = %d %f\n", n, Volt[n]);
	//for (int n = 0; n < NEpsr; n++) printf("epsr = %d %f\n", n, Epsr[n]);
	//for (int n = 0; n < NGeom; n++) printf("geometry = %d %d %d %d %f %f %f %f %f %f\n", n, Geom[n].type, Geom[n].pid, Geom[n].shape, Geom[n].g[0], Geom[n].g[1], Geom[n].g[2], Geom[n].g[3], Geom[n].g[4], Geom[n].g[5]);
	//for (int n = 0; n < NL1d; n++) printf("1d = %d %c %c %f %f\n", n, L1d[n].component, L1d[n].direction, pos1d1[n], pos1d2[n]);
	//for (int n = 0; n < NP2d; n++) printf("2d = %d %c %c %f\n", n, P2d[n].component, P2d[n].direction, pos2d0[n]);
	//printf("solver = %f %d %d %e\n", Solver.omega, Solver.maxiter, Solver.nout, Solver.converg);

	// error check

	if (check) {
		if (nxr <= 0) {
			fprintf(stderr, "*** no xmesh data\n");
			return 1;
		}
		if (nyr <= 0) {
			fprintf(stderr, "*** no ymesh data\n");
			return 1;
		}
		if (nzr <= 0) {
			fprintf(stderr, "*** no zmesh data\n");
			return 1;
		}
		for (int i = 0; i < nxr; i++) {
			if ((xr[i] > xr[i + 1]) || (dxr[i] <= 0)) {
				fprintf(stderr, "*** invalid xmesh data\n");
				return 1;
			}
		}
		for (int j = 0; j < nyr; j++) {
			if ((yr[j] > yr[j + 1]) || (dyr[j] <= 0)) {
				fprintf(stderr, "*** invalid ymesh data\n");
				return 1;
			}
		}
		for (int k = 0; k < nzr; k++) {
			if ((zr[k] > zr[k + 1]) || (dzr[k] <= 0)) {
				fprintf(stderr, "*** invalid zmesh data\n");
				return 1;
			}
		}
		if (NVolt < 3) {
			fprintf(stderr, "*** number of volt data < 2\n");
			return 1;
		}
		if ((Solver.maxiter <= 0) || (Solver.nout <= 0)) {
			fprintf(stderr, "*** invalid solver data\n");
			return 1;
		}
	}

	for (int n = 0; n < NEpsr; n++) {
		if (Epsr[n] <= 0) {
			fprintf(stderr, "*** invalid epsr data #%d\n", n + 1);
			return 1;
		}
	}

	for (int n = 0; n < NGeom; n++) {
		if ((Geom[n].type < 1) || (Geom[n].type > 2)) {
			fprintf(stderr, "*** invalid type of geometry data #%d\n", n + 1);
			return 1;
		}
		if (Geom[n].pid < 1) {
			fprintf(stderr, "*** invalid property of geometry data #%d\n", n + 1);
			return 1;
		}
		if (Geom[n].shape < 1) {
			fprintf(stderr, "*** invalid shape of geometry data #%d\n", n + 1);
			return 1;
		}
		if ((Geom[n].type == 1) && (Geom[n].pid > NVolt)) {
			fprintf(stderr, "*** invalid volt of geometry data #%d\n", n + 1);
			return 1;
		}
		if ((Geom[n].type == 2) && (Geom[n].pid > NEpsr)) {
			fprintf(stderr, "*** invalid epsr of geometry data #%d\n", n + 1);
			return 1;
		}
	}
	for (int n = 0; n < NL1d; n++) {
		if ((L1d[n].direction != 'X') && (L1d[n].direction != 'Y') && (L1d[n].direction != 'Z')) {
			fprintf(stderr, "*** invalid direction of plot1d #%d\n", n + 1);
			return 1;
		}
	}
	for (int n = 0; n < NP2d; n++) {
		if ((P2d[n].direction != 'X') && (P2d[n].direction != 'Y') && (P2d[n].direction != 'Z')) {
			fprintf(stderr, "*** invalid direction of plot2d #%d\n", n + 1);
			return 1;
		}
	}

	// setup
	if (runMode < 2) {
		// number of cells
		setup_cells(nxr, nyr, nzr, dxr, dyr, dzr);

		// node
		setup_node(nxr, nyr, nzr, xr, yr, zr, dxr, dyr, dzr);

		// fit geometry without thickness
		fitgeometry();

		// 1d
		setup_1d();

		// 2d
		setup_2d();
	}

	// free
	free(xr);
	free(yr);
	free(zr);
	free(dxr);
	free(dyr);
	free(dzr);

	return 0;
}
