/*
monitor.c
*/

#include "ost.h"

// CPU memory size [MB]
static int memory_size_cpu(int vector, int commsize)
{
	size_t n = (size_t)(Nx + 3) * (Ny + 3) * (Nz + 3) / commsize;
	size_t size
		= 1 * n * sizeof(real_t)       // V
		+ 2 * n * sizeof(id_t)         // idVolt, idEpsr
		+ NGeom * sizeof(geometry_t)   // Geom
		+ (vector == 1 ? n * (sizeof(real_t) + sizeof(id_v_t)) : 0);  // Epsr_v, idVolt_v

	return (int)(size / 1024 / 1024) + 1;
}

// GPU memory size [MB]
static int memory_size_gpu(int commsize)
{
	size_t n = (size_t)(Nx + 3) * (Ny + 3) * (Nz + 3) / commsize;
	size_t size
		= 1 * n * sizeof(real_t)   // V
		+ 2 * n * sizeof(id_t);    // idVolt, idEpsr

	return (int)(size / 1024 / 1024) + 1;
}

// output filesize [MB]
static int output_size()
{
	size_t n = ((size_t)(Nx + 1)) * (Ny + 1) * (Nz + 1);
	size_t size
		= 1 * n * sizeof(real_t)   // V
		+ 2 * n * sizeof(id_t);    // idVolt, idEpsr

	return (int)(size / 1024 / 1024) + 1;
}

// title or message
static void monitor1_(FILE *fp, const char msg[])
{
	fprintf(fp, "%s\n", msg);
	fflush(fp);
}

// condition
static void monitor2_(FILE *fp, int gpu, int vector, int commsize)
{
	time_t now;
	time(&now);

	int nv = 0, ne = 0;
	for (int n = 0; n < NGeom; n++) {
		if (Geom[n].type == 1) nv++;
		if (Geom[n].type == 2) ne++;
	}

	fprintf(fp, "%s", ctime(&now));
	fprintf(fp, "Title = %s\n", Title);
	fprintf(fp, "Cells = %d x %d x %d = %zd\n", Nx, Ny, Nz, (int64_t)Nx * Ny * Nz);
	fprintf(fp, "No. of voltages    = %d\n", NVolt - 1);
	fprintf(fp, "No. of dielecrics  = %d\n", NEpsr - 1);
	fprintf(fp, "No. of geometries  = %d + %d = %d\n", nv, ne, NGeom);
	if (runMode == 0) fprintf(fp, "No. of plot 1d.s   = %d\n", NL1d);
	if (runMode == 0) fprintf(fp, "No. of plot 2d.s   = %d\n", NP2d);
	fprintf(fp, "CPU Memory %s = %d [MB]\n", (commsize > 1 ? "/ proc." : "       "), memory_size_cpu((vector == 1) & !gpu, commsize));
	if (gpu) fprintf(fp, "GPU Memory %s = %d [MB]\n", (commsize > 1 ? "/ proc." : "       "), memory_size_gpu(commsize));
	if (runMode == 1) fprintf(fp, "Output filesize    = %d [MB]\n", output_size());
	fprintf(fp, "Omega              = %g\n", Solver.omega);
	fprintf(fp, "Max iterations     = %d\n", Solver.maxiter);
	fprintf(fp, "Convergence = %.7f\n", Solver.converg);
	fprintf(fp, "=== iteration start ===\n");
	fprintf(fp, "      step    residual\n");
	fflush(fp);
}

// charge and energy
static void monitor3_(FILE *fp)
{
	fprintf(fp, "%s\n", "Volt#   Q[C]");
	for (int n = 1; n < NVolt; n++) {
		fprintf(fp, "%5d %13.5e\n", n, Echar[n]);
	}
	fprintf(fp, "Total energy [W] = %.5e\n", Echar[0]);
	fflush(fp);
}

// output files
static void monitor4_(FILE *fp, int ev)
{
	fprintf(fp, "=== output files ===\n");
	fprintf(fp, "%s, %s", FN_log, (ev ? FN_geom3d_1 : FN_geom3d_0));
	if      (runMode == 0) {
		fprintf(fp, ", %s, %s", (ev ? FN_ev2d_1 : FN_ev2d_0), (ev ? FN_ev3d_1 : FN_ev3d_0));
		if (NL1d && L1dLog) {
			fprintf(fp, ", %s", FN_1d);
		}
		if (NP2d && P2dLog) {
			fprintf(fp, ", %s", FN_2d);
		}
	}
	else if (runMode == 1) {
		fprintf(fp, ", %s", FN_out);
	}
	fprintf(fp, "\n");
	fflush(fp);
}

// cpu time
static void monitor5_(FILE *fp, const double cpu[])
{
	time_t now;
	time(&now);

	fprintf(fp, "%s\n", "=== normal end ===");
	fprintf(fp, "%s", ctime(&now));
	fprintf(fp, "%s\n", "=== cpu time [sec] ===");
	fprintf(fp, "part-1 : %10.3f\n", cpu[0]);
	fprintf(fp, "part-2 : %10.3f\n", cpu[1]);
	fprintf(fp, "part-3 : %10.3f\n", cpu[2]);
	fprintf(fp, "%s\n", "---------------------");
	fprintf(fp, "total  : %10.3f\n", cpu[0] + cpu[1] + cpu[2]);
	fflush(fp);
}

void monitor1(FILE *fp, const char msg[])
{
	monitor1_(fp,     msg);
	monitor1_(stdout, msg);
}

void monitor2(FILE *fp, int gpu, int vector, int commsize)
{
	monitor2_(fp,     gpu, vector, commsize);
	monitor2_(stdout, gpu, vector, commsize);
}

void monitor3(FILE *fp)
{
	monitor3_(fp);
	monitor3_(stdout);
}

void monitor4(FILE *fp, int ev)
{
	monitor4_(fp,     ev);
	monitor4_(stdout, ev);
}

void monitor5(FILE *fp, const double cpu[])
{
	monitor5_(fp,     cpu);
	monitor5_(stdout, cpu);
}
