/*
sor.c

SOR method
*/

#include "ost.h"
#include "ost_prototype.h"

// debug
static real_t update_serial(void);
static real_t update_neumann(void);

// red-black (novector)
static real_t update_rb_novector(int oe)
{
	const real_t omega = (real_t)Solver.omega;

	const int64_t ni = D(1, 0, 0) - D(0, 0, 0);
	const int64_t nj = D(0, 1, 0) - D(0, 0, 0);
	const int64_t nk = D(0, 0, 1) - D(0, 0, 0);

	real_t res2 = 0;
	int64_t i;
#ifdef _OPENMP
#pragma omp parallel for reduction (+: res2)
#endif
	for (        i = 0; i <= Nx; i++) {
	for (int64_t j = 0; j <= Ny; j++) {
		const int64_t kmin = ((i + j) % 2 == oe) ? 0 : 1;
		const int64_t kmax = Nz;
#ifdef __NEC__
#pragma _NEC ivdep
#elif __FUJITSU
#pragma loop novrec
#endif
		for (int64_t k = kmin; k <= kmax; k += 2) {
			const int64_t n = D(i, j, k);
			const real_t e = Epsr[idEpsr[n]];

			const real_t axp = (e + Epsr[idEpsr[n + ni]]) * RXp[i];
			const real_t axm = (e + Epsr[idEpsr[n - ni]]) * RXm[i];
			const real_t ayp = (e + Epsr[idEpsr[n + nj]]) * RYp[j];
			const real_t aym = (e + Epsr[idEpsr[n - nj]]) * RYm[j];
			const real_t azp = (e + Epsr[idEpsr[n + nk]]) * RZp[k];
			const real_t azm = (e + Epsr[idEpsr[n - nk]]) * RZm[k];

			const real_t res = (idVolt[n] ? 0 : 1) * omega * ((
				axp * V[n + ni] +
				axm * V[n - ni] +
				ayp * V[n + nj] +
				aym * V[n - nj] +
				azp * V[n + nk] +
				azm * V[n - nk]) / (axp + axm + ayp + aym + azp + azm) - V[n]);

			V[n] += res;
			res2 += res * res;
		}
	}
	}

	return res2;
}

// red-black (vector)
static real_t update_rb_vector(int oe)
{
	const real_t omega = (real_t)Solver.omega;

	const int64_t ni = D(1, 0, 0) - D(0, 0, 0);
	const int64_t nj = D(0, 1, 0) - D(0, 0, 0);
	const int64_t nk = D(0, 0, 1) - D(0, 0, 0);

	real_t res2 = 0;
	int64_t i;
#ifdef _OPENMP
#pragma omp parallel for reduction (+: res2)
#endif
	for (        i = 0; i <= Nx; i++) {
	for (int64_t j = 0; j <= Ny; j++) {
		const int64_t kmin = ((i + j) % 2 == oe) ? 0 : 1;
		const int64_t kmax = Nz;
		const int64_t n = D(i, j, kmin);
		real_t *ptr = &V[n];
		real_t *ptr_vxp = &V[n + ni];
		real_t *ptr_vxm = &V[n - ni];
		real_t *ptr_vyp = &V[n + nj];
		real_t *ptr_vym = &V[n - nj];
		real_t *ptr_vzp = &V[n + nk];
		real_t *ptr_vzm = &V[n - nk];
		real_t *ptr_e   = &Epsr_v[n];
		real_t *ptr_exp = &Epsr_v[n + ni];
		real_t *ptr_exm = &Epsr_v[n - ni];
		real_t *ptr_eyp = &Epsr_v[n + nj];
		real_t *ptr_eym = &Epsr_v[n - nj];
		real_t *ptr_ezp = &Epsr_v[n + nk];
		real_t *ptr_ezm = &Epsr_v[n - nk];
		real_t rxp = RXp[i];
		real_t rxm = RXm[i];
		real_t ryp = RYp[j];
		real_t rym = RYm[j];
		real_t *ptr_rzp = &RZp[kmin];
		real_t *ptr_rzm = &RZm[kmin];
		id_v_t *ptr_id = &idVolt_v[n];
#ifdef __NEC__
#pragma _NEC ivdep
#elif __FUJITSU
#pragma loop novrec
#endif
		for (int64_t k = kmin; k <= kmax; k += 2) {
			const real_t e = *ptr_e;

			const real_t axp = (e + (*ptr_exp)) * rxp;
			const real_t axm = (e + (*ptr_exm)) * rxm;
			const real_t ayp = (e + (*ptr_eyp)) * ryp;
			const real_t aym = (e + (*ptr_eym)) * rym;
			const real_t azp = (e + (*ptr_ezp)) * (*ptr_rzp);
			const real_t azm = (e + (*ptr_ezm)) * (*ptr_rzm);

			const real_t res = (*ptr_id ? 0 : 1) * omega * ((
				axp * (*ptr_vxp) +
				axm * (*ptr_vxm) +
				ayp * (*ptr_vyp) +
				aym * (*ptr_vym) +
				azp * (*ptr_vzp) +
				azm * (*ptr_vzm)) / (axp + axm + ayp + aym + azp + azm) - (*ptr));

			*ptr += res;
			res2 += res * res;

			ptr     += 2;
			ptr_vxp += 2;
			ptr_vxm += 2;
			ptr_vyp += 2;
			ptr_vym += 2;
			ptr_vzp += 2;
			ptr_vzm += 2;
			ptr_e   += 2;
			ptr_exp += 2;
			ptr_exm += 2;
			ptr_eyp += 2;
			ptr_eym += 2;
			ptr_ezp += 2;
			ptr_ezm += 2;
			ptr_id  += 2;
			ptr_rzp += 2;
			ptr_rzm += 2;
		}
	}
	}

	return res2;
}

void sor(int vector, FILE *fp)
{
	char str[BUFSIZ];
	const char msg1[] = "  --- converged.";
	const char msg2[] = "  --- diverged.";
	const char msg3[] = "  --- not converged.";
	const char fmt[] = "   %7d    %.7f";

	const int64_t num = (Nx + 3) * (Ny + 3) * (Nz + 3);

	// V
	V = (real_t *)malloc(num * sizeof(real_t));
	memset(V, 0, num * sizeof(real_t));

	// vector
	if (vector > 0) {
		// Er array
		Epsr_v = (real_t *)malloc(num * sizeof(real_t));
		for (int64_t n = 0; n < num; n++) {
			Epsr_v[n] = Epsr[idEpsr[n]];  // direct access
		}
		// Volt array
		idVolt_v = (id_v_t *)malloc(num * sizeof(id_v_t));
		for (int64_t n = 0; n < num; n++) {
			idVolt_v[n] = idVolt[n];  // id_t -> int32/64
		}
	}

	// residual
	size_t r_size = (Solver.maxiter + 1) * sizeof(double);
	Residual = (double *)malloc(r_size);
	memset(Residual, 0, r_size);

	// electrode
	electrode(0, Nx, 0, Ny, 0, Nz);

	// initial field
	real_t vmin, vmax;
	initial(&vmin, &vmax, 0, Nx, 0, Ny, 0, Nz);

	// iteration
	int iter;
	for (iter = 0; iter <= Solver.maxiter; iter++) {
		// update
		real_t res2 = 0;
		if      (vector ==  1) {
			res2 = update_rb_vector(0)
			     + update_rb_vector(1);
		}
		else if (vector ==  0) {
			res2 = update_rb_novector(0)
			     + update_rb_novector(1);
		}
		else if (vector == -1) {
			res2 = update_serial();
		}
		else if (vector == -2) {
			res2 = update_neumann();
		}

		// residual
		Residual[iter] = sqrt(res2 / ((double)(Nx + 1) * (Ny + 1) * (Nz + 1)));

		// monitor
		if ((iter % Solver.nout) == 0 || (Residual[iter] < Solver.converg)) {
			sprintf(str, fmt, iter, Residual[iter]);
			monitor1(fp, str);
		}

		// converged / diverged / not converged
		if (Residual[iter] < Solver.converg) {
			monitor1(fp, msg1);
			break;
		}
		if (Residual[iter] > 1e10) {
			monitor1(fp, msg2);
			break;
		}
		if (iter == Solver.maxiter) {
			monitor1(fp, msg3);
			break;
		}
	}

	// iterations
	NumIter = iter + 1;

	// final
	final(vmin, vmax, 0, Nx, 0, Ny, 0, Nz);

	// edge
	edge(0, Nx, 0, Ny, 0, Nz);

	// vector (free)
	if (vector) {
		if (Epsr_v   != NULL) free(Epsr_v);
		if (idVolt_v != NULL) free(idVolt_v);
	}
}

// non red-black (debug, serial)
static real_t update_serial(void)
{
	const real_t omega = (real_t)Solver.omega;

	const int64_t ni = D(1, 0, 0) - D(0, 0, 0);
	const int64_t nj = D(0, 1, 0) - D(0, 0, 0);
	const int64_t nk = D(0, 0, 1) - D(0, 0, 0);

	real_t res2 = 0;
	for (int64_t i = 0; i <= Nx; i++) {
	for (int64_t j = 0; j <= Ny; j++) {
	for (int64_t k = 0; k <= Nz; k++) {
		const int64_t n = D(i, j, k);
		const real_t e = Epsr[idEpsr[n]];

		const real_t axp = (e + Epsr[idEpsr[n + ni]]) * RXp[i];
		const real_t axm = (e + Epsr[idEpsr[n - ni]]) * RXm[i];
		const real_t ayp = (e + Epsr[idEpsr[n + nj]]) * RYp[j];
		const real_t aym = (e + Epsr[idEpsr[n - nj]]) * RYm[j];
		const real_t azp = (e + Epsr[idEpsr[n + nk]]) * RZp[k];
		const real_t azm = (e + Epsr[idEpsr[n - nk]]) * RZm[k];

		const real_t res = (idVolt[n] ? 0 : 1) * omega * ((
			axp * V[n + ni] +
			axm * V[n - ni] +
			ayp * V[n + nj] +
			aym * V[n - nj] +
			azp * V[n + nk] +
			azm * V[n - nk]) / (axp + axm + ayp + aym + azp + azm) - V[n]);

		V[n] += res;
		res2 += res * res;
	}
	}
	}

	return res2;
}

// non red-black (debug, Neumann BC)
static real_t update_neumann(void)
{
	const real_t omega = (real_t)Solver.omega;

	const int64_t ni = D(1, 0, 0) - D(0, 0, 0);
	const int64_t nj = D(0, 1, 0) - D(0, 0, 0);
	const int64_t nk = D(0, 0, 1) - D(0, 0, 0);

	real_t res2 = 0;
	for (int i = 0; i <= Nx; i++) {
	for (int j = 0; j <= Ny; j++) {
	for (int k = 0; k <= Nz; k++) {
		const int64_t n = D(i, j, k);
		if (!idVolt[n]) {
			if      (i == 0) {
				V[n] = V[n + ni];
			}
			else if (i == Nx) {
				V[n] = V[n - ni];
			}
			else if (j == 0) {
				V[n] = V[n + nj];
			}
			else if (j == Ny) {
				V[n] = V[n - nj];
			}
			else if (k == 0) {
				V[n] = V[n + nk];
			}
			else if (k == Nz) {
				V[n] = V[n - nk];
			}
			else {
				const real_t e = Epsr[idEpsr[n]];

				const real_t axp = (e + Epsr[idEpsr[n + ni]]) * RXp[i];
				const real_t axm = (e + Epsr[idEpsr[n - ni]]) * RXm[i];
				const real_t ayp = (e + Epsr[idEpsr[n + nj]]) * RYp[j];
				const real_t aym = (e + Epsr[idEpsr[n - nj]]) * RYm[j];
				const real_t azp = (e + Epsr[idEpsr[n + nk]]) * RZp[k];
				const real_t azm = (e + Epsr[idEpsr[n - nk]]) * RZm[k];

				const real_t res = omega * ((
					axp * V[n + ni] +
					axm * V[n - ni] +
					ayp * V[n + nj] +
					aym * V[n - nj] +
					azp * V[n + nk] +
					azm * V[n - nk]) / (axp + axm + ayp + aym + azp + azm) - V[n]);

				V[n] += res;
				res2 += res * res;
			}
		}
	}
	}
	}

	return res2;
}
